@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -181,8 +181,10 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
+	{
 		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
 			std::endl;
+	}
 
 	// read the key file
 	std::string armored_seckey, thispeer = peers[0];
@@ -330,12 +332,12 @@ int main
 	// show information w.r.t. primary key
 	std::string kid, fpr;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
-		KeyidCompute(prv->pub->pub_hashing, kid);
+		KeyidConvert(prv->pub->id, kid);
 	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintComputePretty(prv->pub->pub_hashing, fpr);
-	std::cout << "OpenPGP V4 Key ID of primary key: " << std::endl << "\t";
+		FingerprintConvertPretty(prv->pub->fingerprint, fpr);
+	std::cout << "OpenPGP Key ID of primary key: " << std::endl << "\t";
 	std::cout << kid << std::endl;
-	std::cout << "OpenPGP V4 fingerprint of primary key: " << std::endl << "\t";
+	std::cout << "OpenPGP Fingerprint of primary key: " << std::endl << "\t";
 	std::cout << fpr << std::endl;
 	std::cout << "OpenPGP Key Creation Time: " <<
 		std::endl << "\t" << ctime(&prv->pub->creationtime);
@@ -447,12 +449,12 @@ int main
 		TMCG_OpenPGP_PrivateSubkey *sub = prv->private_subkeys[0];
 		std::string kid2, fpr2;
 		CallasDonnerhackeFinneyShawThayerRFC4880::
-			KeyidCompute(sub->pub->sub_hashing, kid2);
+			KeyidConvert(sub->pub->id, kid2);
 		CallasDonnerhackeFinneyShawThayerRFC4880::
-			FingerprintComputePretty(sub->pub->sub_hashing, fpr2);
-		std::cout << "OpenPGP V4 Key ID of subkey: " << std::endl << "\t";
+			FingerprintConvertPretty(sub->pub->fingerprint, fpr2);
+		std::cout << "OpenPGP Key ID of subkey: " << std::endl << "\t";
 		std::cout << kid2 << std::endl;
-		std::cout << "OpenPGP V4 fingerprint of subkey: " << std::endl << "\t";
+		std::cout << "OpenPGP Fingerprint of subkey: " << std::endl << "\t";
 		std::cout << fpr2 << std::endl;
 		std::cout << "OpenPGP Key Creation Time: " <<
 			std::endl << "\t" << ctime(&sub->pub->creationtime);
@@ -479,35 +481,35 @@ int main
 		}
 		if (sub->pub->revkeys.size() == 0)
 			std::cout << "\t" << "none" << std::endl;
-		size_t allflags = sub->pub->AccumulateFlags();
+		size_t allflags_sub = sub->pub->AccumulateFlags();
 		std::cout << "OpenPGP Key Flags: " << std::endl << "\t";
 		// The key may be used to certify other keys.
-		if ((allflags & 0x01) == 0x01)
+		if ((allflags_sub & 0x01) == 0x01)
 			std::cout << "C";
 		// The key may be used to sign data.
-		if ((allflags & 0x02) == 0x02)
+		if ((allflags_sub & 0x02) == 0x02)
 			std::cout << "S";
 		// The key may be used encrypt communications.
-		if ((allflags & 0x04) == 0x04)
+		if ((allflags_sub & 0x04) == 0x04)
 			std::cout << "E";
 		// The key may be used encrypt storage.
-		if ((allflags & 0x08) == 0x08)
+		if ((allflags_sub & 0x08) == 0x08)
 			std::cout << "e";
 		// The private component of this key may have
 		// been split by a secret-sharing mechanism.
-		if ((allflags & 0x10) == 0x10)
+		if ((allflags_sub & 0x10) == 0x10)
 			std::cout << "D";
 		// The key may be used for authentication.
-		if ((allflags & 0x20) == 0x20)
+		if ((allflags_sub & 0x20) == 0x20)
 			std::cout << "A";
 		// The private component of this key may be
 		// in the possession of more than one person.
-		if ((allflags & 0x80) == 0x80)
+		if ((allflags_sub & 0x80) == 0x80)
 			std::cout << "G";
 		// This key may be used for timestamping. [draft RFC 4880bis]
-		if ((allflags & 0x0800) == 0x0800)
+		if ((allflags_sub & 0x0800) == 0x0800)
 			std::cout << "T";
-		if (allflags == 0x00)
+		if (allflags_sub == 0x00)
 			std::cout << "undefined";
 		std::cout << std::endl;
 		// show information w.r.t. tElG
@@ -705,6 +707,8 @@ int main
 	if (dss != NULL)
 		delete dss;
 	delete prv;
+
+	// unlock secure memory
 	if (should_unlock)
 		unlock_memory();
 
