@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -57,13 +57,16 @@ static const char *protocol = "DKGPG-keysign-1.0";
 #include "dkg-common.hh"
 
 int								pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int								self_pipefd[2];
 int								broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int								broadcast_self_pipefd[2];
 pid_t							pid[DKGPG_MAX_N];
 std::vector<std::string>		peers;
 bool							instance_forked = false;
 
 tmcg_openpgp_secure_string_t	passphrase;
 std::string						ifilename, ofilename, kfilename;
+std::string						Kfilename, fingerprint;
 std::string						passwords, hostname, port, URI, u, yfilename;
 
 int 							opt_verbose = 0;
@@ -74,10 +77,12 @@ char							*opt_hostname = NULL;
 char							*opt_URI = NULL;
 char							*opt_u = NULL;
 char							*opt_k = NULL;
+char							*opt_K = NULL;
+char							*opt_fingerprint = NULL;
 char							*opt_y = NULL;
 unsigned long int				opt_e = 0, opt_p = 55000, opt_W = 5;
-bool							opt_r = false;
 bool							opt_1 = false, opt_2 = false, opt_3 = false;
+bool							opt_a = false, opt_r = false;
 
 void run_instance
 	(size_t whoami, const time_t sigtime, const time_t sigexptime,
@@ -103,9 +108,43 @@ void run_instance
 		exit(-1);
 
 	// read the public key
+	bool parse_ok;
 	std::string armored_pubkey;
-	if (!read_key_file(opt_ifilename, armored_pubkey))
-		exit(-1);
+	if (opt_ifilename != NULL)
+	{
+		if (!read_key_file(opt_ifilename, armored_pubkey))
+			exit(-1);
+	}
+	else if (opt_K != NULL)
+	{
+		std::string armored_certring;
+		if (!read_key_file(Kfilename, armored_certring))
+			exit(-1);
+		TMCG_OpenPGP_Keyring *certring = NULL;
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PublicKeyringParse(armored_certring, opt_verbose_ring, certring);
+		if (!parse_ok)
+		{
+			std::cerr << "ERROR: cannot use the given keyring" << std::endl;
+			exit(-1);
+		}
+		if (!get_key_by_fingerprint(certring, fingerprint, opt_verbose,
+			armored_pubkey))
+		{
+			if (!get_key_by_keyid(certring, fingerprint, opt_verbose,
+				armored_pubkey))
+			{
+				std::cerr << "ERROR: public key not found in keyring" <<
+					std::endl;
+				delete certring;
+				exit(-1);
+			}
+		}
+		delete certring;
+	}
 
 	// read the keyring
 	std::string armored_pubring;
@@ -118,11 +157,13 @@ void run_instance
 	// parse the keyring, the private key and corresponding signatures
 	TMCG_OpenPGP_Prvkey *prv = NULL;
 	TMCG_OpenPGP_Keyring *ring = NULL;
-	bool parse_ok;
 	if (opt_k)
 	{
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
-			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+			PublicKeyringParse(armored_pubring, opt_verbose_ring, ring);
 		if (!parse_ok)
 		{
 			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
@@ -214,6 +255,7 @@ void run_instance
 		delete prv;
 		exit(-1);
 	}
+	delete ring;
 
 	// initialize signature scheme
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = NULL;
@@ -229,7 +271,6 @@ void run_instance
 		{
 			delete dss;
 			delete primary;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -240,7 +281,6 @@ void run_instance
 				std::endl;
 			delete dss;
 			delete primary;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -255,12 +295,11 @@ void run_instance
 				std::string pwd;
 				if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 				{
-					std::cerr << "ERROR: S_" << whoami << ": " <<
-						"cannot read password for protecting channel to S_" <<
+					std::cerr << "ERROR: p_" << whoami << ": " <<
+						"cannot read password for protecting channel to p_" <<
 						i << std::endl;
 					delete dss;
 					delete primary;
-					delete ring;
 					delete prv;
 					exit(-1);
 				}
@@ -268,25 +307,30 @@ void run_instance
 				if (((i + 1) < peers.size()) &&
 					!TMCG_ParseHelper::nx(passwords, '/'))
 				{
-					std::cerr << "ERROR: S_" << whoami << ": " <<
+					std::cerr << "ERROR: p_" << whoami << ": " <<
 						"cannot skip to next password for protecting channel" <<
-						" to S_" << (i + 1) << std::endl;
+						" to p_" << (i + 1) << std::endl;
 					delete dss;
 					delete primary;
-					delete ring;
 					delete prv;
 					exit(-1);
 				}
 			}
 			else
 			{
-				// simple key -- we assume that GNUnet will provide secure channels
-				key << "dkg-keysign::S_" << (i + whoami);
+				// simple key -- we assume that GNUnet provides secure channels
+				key << "dkg-keysign::p_" << (i + whoami);
 			}
-			uP_in.push_back(pipefd[i][whoami][0]);
+			if (i == whoami)
+				uP_in.push_back(self_pipefd[0]);
+			else
+				uP_in.push_back(pipefd[i][whoami][0]);
 			uP_out.push_back(pipefd[whoami][i][1]);
 			uP_key.push_back(key.str());
-			bP_in.push_back(broadcast_pipefd[i][whoami][0]);
+			if (i == whoami)
+				bP_in.push_back(broadcast_self_pipefd[0]);
+			else
+				bP_in.push_back(broadcast_pipefd[i][whoami][0]);
 			bP_out.push_back(broadcast_pipefd[whoami][i][1]);
 			bP_key.push_back(key.str());
 		}
@@ -308,82 +352,18 @@ void run_instance
 			aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 		rbc->setID(myID);
 		// perform a simple exchange test with debug output
-		for (size_t i = 0; i < num_xtests; i++)
-		{
-			mpz_t xtest;
-			mpz_init_set_ui(xtest, i);
-			std::cerr << "S_" << whoami << ": xtest = " << xtest << " <-> ";
-			rbc->Broadcast(xtest);
-			for (size_t ii = 0; ii < peers.size(); ii++)
-			{
-				if (!rbc->DeliverFrom(xtest, ii))
-					std::cerr << "<X> ";
-				else
-					std::cerr << xtest << " ";
-			}
-			std::cerr << std::endl;
-			mpz_clear(xtest);
-		}
+		xtest(num_xtests, whoami, peers.size(), rbc);
 		// participants must agree on a common signature creation time (OpenPGP)
-		if (opt_verbose)
-			std::cerr << "INFO: agree on a signature creation time for" <<
-				" OpenPGP" << std::endl;
-		std::vector<time_t> tvs;
-		mpz_t mtv;
-		mpz_init_set_ui(mtv, sigtime);
-		rbc->Broadcast(mtv);
-		tvs.push_back(sigtime);
-		for (size_t i = 0; i < peers.size(); i++)
-		{
-			if (i != whoami)
-			{
-				if (rbc->DeliverFrom(mtv, i))
-				{
-					time_t utv;
-					utv = (time_t)mpz_get_ui(mtv);
-					tvs.push_back(utv);
-				}
-				else
-				{
-					std::cerr << "WARNING: S_" << whoami << ": no signature" <<
-						" creation time stamp received from " << i << std::endl;
-				}
-			}
-		}
-		mpz_clear(mtv);
-		std::sort(tvs.begin(), tvs.end());
-		if (tvs.size() < (peers.size() - T_RBC))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": not enough timestamps" <<
-				" received" << std::endl;
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete primary;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		// use a median value as some kind of gentle agreement
-		csigtime = tvs[tvs.size()/2];
-		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": canonicalized signature" <<
-				" creation time = " << csigtime << std::endl;
+		csigtime = agree_time(sigtime, whoami, peers.size(), opt_verbose, rbc);
 		// select hash algorithm for OpenPGP based on |q| (size in bit)
-		if (mpz_sizeinbase(dss->q, 2L) == 256)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-		else if (mpz_sizeinbase(dss->q, 2L) == 384)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-		else if (mpz_sizeinbase(dss->q, 2L) == 512)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
-		else
+		if (!select_hashalgo(dss, hashalgo))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": selecting hash" <<
+			std::cerr << "ERROR: p_" << whoami << ": selecting hash" <<
 				" algorithm failed for |q| = " << mpz_sizeinbase(dss->q, 2L) <<
 				std::endl;
 			delete rbc, delete aiou, delete aiou2;
 			delete dss;
 			delete primary;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -397,38 +377,41 @@ void run_instance
 	// prepare the fingerprint, the trailer, and the accumulator of the
 	// certification (revocation) signatures
 	std::string fpr;
-	tmcg_openpgp_octets_t trailer, acc, issuer;
+	tmcg_openpgp_octets_t trailer, acc;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintComputePretty(primary->pub_hashing, fpr);
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(prv->pub->pub_hashing, issuer);
+		FingerprintConvertPretty(primary->fingerprint, fpr);
 	if (opt_y == NULL)
 	{
 		if (opt_r)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CERTIFICATION_REVOCATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_1)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_PERSONA_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_2)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CASUAL_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_3)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_GENERIC_CERTIFICATION,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 	}
 	else
 	{
@@ -436,31 +419,37 @@ void run_instance
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CERTIFICATION_REVOCATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_1)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_PERSONA_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_2)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_CASUAL_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else if (opt_3)
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 		else
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareCertificationSignature(
 					TMCG_OPENPGP_SIGNATURE_GENERIC_CERTIFICATION, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
+					trailer);
 	}
 	acc.insert(acc.end(), primary->packet.begin(), primary->packet.end());
 
-	// loop through all or selected valid user IDs	
+	// loop through all or selected valid user IDs
+	bool anything_signed = false;
 	for (size_t j = 0; j < primary->userids.size(); j++)
 	{
 		// user ID not selected?
@@ -480,178 +469,32 @@ void run_instance
 			std::cerr << "INFO: going to sign user ID \"" <<
 				primary->userids[j]->userid_sanitized << "\" of key with" <<
 				" fingerprint [ " << fpr << "]" << std::endl;
-
-		// prepare the hash value
-		tmcg_openpgp_byte_t buffer[1024];
-		size_t buflen = 0;
-		memset(buffer, 0, sizeof(buffer));
-		if (opt_verbose > 1)
-			std::cerr << std::hex << "INFO: hash = ";
-		for (size_t i = 0; i < hash.size(); i++, buflen++)
+		if (opt_a)
 		{
-			if (i < sizeof(buffer))
-				buffer[i] = hash[i];
-			if (opt_verbose > 1)
-				std::cerr << (int)hash[i] << " ";
+			std::string p = "Please confirm operation by entering the string";
+			if (!check_confirmation(p))
+			{
+				std::cerr << "WARNING: operation has been cancelled" <<
+					std::endl;
+				continue;
+			}
 		}
-		if (opt_verbose > 1)
-			std::cerr << std::dec << std::endl;
+		anything_signed = true;
 
 		// sign the hash value
-		gcry_error_t ret;
-		gcry_mpi_t r, s;
-		r = gcry_mpi_new(2048);
-		s = gcry_mpi_new(2048);
-		if (opt_y == NULL)
-		{
-			gcry_mpi_t h;
-			ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-			if (ret)
-			{
-				std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan()" <<
-					" failed for h" << std::endl;
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-			if (opt_verbose > 1)
-				std::cerr << "INFO: S_" << whoami << ": h = " << h << std::endl;
-			mpz_t dsa_m, dsa_r, dsa_s;
-			mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
-			if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
-			{
-				std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
-					" failed for dsa_m" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-			gcry_mpi_release(h);
-			std::stringstream err_log_sign;
-			if (opt_verbose)
-				std::cerr << "INFO: S_" << whoami << ": dss.Sign()" << std::endl;
-			if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s,
-				prv->tdss_idx2dkg, prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
-			{
-				std::cerr << "ERROR: S_" << whoami << ": " <<
-					"tDSS Sign() failed" << std::endl;
-				std::cerr << "ERROR: S_" << whoami << ": log follows " <<
-					std::endl << err_log_sign.str();
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-			if (opt_verbose > 1)
-				std::cerr << "INFO: S_" << whoami << ": log follows " <<
-					std::endl << err_log_sign.str();
-			if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
-			{
-				std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-					" failed for dsa_r" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-			if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
-			{
-				std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-					" failed for dsa_s" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		}
-		else
-		{
-			switch (prv->pkalgo)
-			{
-				case TMCG_OPENPGP_PKALGO_RSA:
-				case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
-					ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-						AsymmetricSignRSA(hash, prv->private_key, hashalgo, s);
-					break;
-				case TMCG_OPENPGP_PKALGO_DSA:
-					ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-						AsymmetricSignDSA(hash, prv->private_key, r, s);
-					break;
-				case TMCG_OPENPGP_PKALGO_ECDSA:
-					ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-						AsymmetricSignECDSA(hash, prv->private_key, r, s);
-					break;
-				case TMCG_OPENPGP_PKALGO_EDDSA:
-					ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-						AsymmetricSignEdDSA(hash, prv->private_key, r, s);
-					break;
-			default:
-					std::cerr << "ERROR: public-key algorithm " <<
-						(int)prv->pkalgo << " not supported" << std::endl;
-					gcry_mpi_release(r), gcry_mpi_release(s);
-					delete primary;
-					delete ring;
-					delete prv;
-					exit(-1);
-			}
-			if (ret)
-			{
-				std::cerr << "ERROR: signing of hash value failed " <<
-					"(rc = " << gcry_err_code(ret) << ", str = " <<
-					gcry_strerror(ret) << ")" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-		}
 		tmcg_openpgp_octets_t sig;
-		switch (prv->pkalgo)
+		if (!sign_hash(hash, trailer, left, whoami, peers.size(), prv, hashalgo,
+			sig, opt_verbose, opt_y, dss, aiou, rbc))
 		{
-			case TMCG_OPENPGP_PKALGO_RSA:
-			case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
-				CallasDonnerhackeFinneyShawThayerRFC4880::
-					PacketSigEncode(trailer, left, s, sig);
-				break;
-			case TMCG_OPENPGP_PKALGO_DSA:
-			case TMCG_OPENPGP_PKALGO_ECDSA:
-			case TMCG_OPENPGP_PKALGO_EDDSA:
-			case TMCG_OPENPGP_PKALGO_EXPERIMENTAL7:
-				CallasDonnerhackeFinneyShawThayerRFC4880::
-					PacketSigEncode(trailer, left, r, s, sig);
-				break;
-			default:
-				std::cerr << "ERROR: public-key algorithm " <<
-					(int)prv->pkalgo << " not supported" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				delete primary;
-				delete ring;
-				delete prv;
-				exit(-1);
+			if (opt_y == NULL)
+			{
+				delete rbc, delete aiou, delete aiou2;
+				delete dss;
+			}
+			delete primary;
+			delete prv;
+			exit(-1);
 		}
-		gcry_mpi_release(r), gcry_mpi_release(s);
 
 		// attach the generated certification (revocation) signature to
 		// public key, selected user IDs, and exportable self-signatures
@@ -672,25 +515,23 @@ void run_instance
 	if (opt_y == NULL)
 	{
 		// at the end: deliver some more rounds for still waiting parties
-		time_t synctime = aiounicast::aio_timeout_long;
+		time_t synctime = (opt_W * 6);
 		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": waiting approximately " <<
+			std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
 				(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
 				std::endl;
+		if (rbc == NULL)
+			exit(-2); // should never happen: only here to make scan-build happy
 		rbc->Sync(synctime);
 		// release RBC
 		delete rbc;
-		// release handles (unicast channel)
+		// release handles (both channels)
 		if (opt_verbose)
 		{
-			std::cerr << "INFO: S_" << whoami << ": unicast channels";
+			std::cerr << "INFO: p_" << whoami << ": unicast channels";
 			aiou->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
-		}
-		// release handles (broadcast channel)
-		if (opt_verbose)
-		{
-			std::cerr << "INFO: S_" << whoami << ": broadcast channel";
+			std::cerr << "INFO: p_" << whoami << ": broadcast channel";
 			aiou2->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
 		}
@@ -700,9 +541,15 @@ void run_instance
 		delete dss;
 	}
 	delete primary;
-	delete ring;
 	delete prv;
 
+	// check the result
+	if (!anything_signed)
+	{
+		std::cerr << "ERROR: no user ID selected and signed" << std::endl;
+		exit(-2);
+	}
+
 	// output the result
 	std::string signedkey;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
@@ -724,6 +571,8 @@ char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 char *gnunet_opt_u = NULL;
 char *gnunet_opt_URI = NULL;
+char *gnunet_opt_K = NULL;
+char *gnunet_opt_fingerprint = NULL;
 char *gnunet_opt_k = NULL;
 char *gnunet_opt_y = NULL;
 unsigned int gnunet_opt_sigexptime = 0;
@@ -731,8 +580,8 @@ unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
 unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_verbose = 0;
-int gnunet_opt_r = 0;
 int gnunet_opt_1 = 0, gnunet_opt_2 = 0, gnunet_opt_3 = 0;
+int gnunet_opt_a = 0, gnunet_opt_r = 0;
 #endif
 
 void fork_instance
@@ -744,7 +593,22 @@ void fork_instance
 	{
 		if (pid[whoami] == 0)
 		{
-			/* BEGIN child code: participant S_i */
+			/* BEGIN child code: participant p_i */
+			if ((self_pipefd[0] == -1) && (self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				self_pipefd[0] = dup(pipefd[whoami][whoami][0]);
+				self_pipefd[1] = dup(pipefd[whoami][whoami][1]);
+			}
+			if ((broadcast_self_pipefd[0] == -1) &&
+				(broadcast_self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				broadcast_self_pipefd[0] =
+					dup(broadcast_pipefd[whoami][whoami][0]);
+				broadcast_self_pipefd[1] =
+					dup(broadcast_pipefd[whoami][whoami][1]);
+			}
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_sigexptime,
@@ -753,9 +617,9 @@ void fork_instance
 			run_instance(whoami, sigtime, opt_e, 0);
 #endif
 			if (opt_verbose)
-				std::cerr << "INFO: S_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: p_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
-			/* END child code: participant S_i */
+			/* END child code: participant p_i */
 		}
 		else
 		{
@@ -792,12 +656,23 @@ int main
 			"issuer has done substantial verification of the claim of identity",
 			&gnunet_opt_3
 		),
+		GNUNET_GETOPT_option_flag('a',
+			"ask",
+			"require confirmation from STDIN for each signature",
+			&gnunet_opt_a
+		),
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
-			"TIME",
+			"INTEGER",
 			"expiration time of generated signature in seconds",
 			&gnunet_opt_sigexptime
 		),
+		GNUNET_GETOPT_option_string('f',
+			"fingerprint",
+			"STRING",
+			"fingerprint of the public key for certification",
+			&gnunet_opt_fingerprint
+		),
 		GNUNET_GETOPT_option_string('H',
 			"hostname",
 			"STRING",
@@ -810,6 +685,12 @@ int main
 			"create certification signature on key resp. user ID from FILENAME",
 			&gnunet_opt_ifilename
 		),
+		GNUNET_GETOPT_option_string('K',
+			"keys",
+			"FILENAME",
+			"select public key for certification from keyring FILENAME",
+			&gnunet_opt_K
+		),
 		GNUNET_GETOPT_option_string('k',
 			"keyring",
 			"FILENAME",
@@ -861,13 +742,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of signing protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -908,6 +789,10 @@ int main
 		opt_URI = gnunet_opt_URI;
 	if (gnunet_opt_u != NULL)
 		opt_u = gnunet_opt_u;
+	if (gnunet_opt_K != NULL)
+		opt_K = gnunet_opt_K;
+	if (gnunet_opt_fingerprint != NULL)
+		opt_fingerprint = gnunet_opt_fingerprint;
 	if (gnunet_opt_k != NULL)
 		opt_k = gnunet_opt_k;
 	if (gnunet_opt_passwords != NULL)
@@ -920,6 +805,10 @@ int main
 		URI = gnunet_opt_URI; // get policy URI from GNUnet options
 	if (gnunet_opt_u != NULL)
 		u = gnunet_opt_u; // get policy URI from GNUnet options
+	if (gnunet_opt_K != NULL)
+		Kfilename = gnunet_opt_K; // get keyring from GNUnet options
+	if (gnunet_opt_fingerprint != NULL)
+		fingerprint = gnunet_opt_fingerprint; // get fingerprint from GNUnet options
 	if (gnunet_opt_k != NULL)
 		kfilename = gnunet_opt_k; // get keyring from GNUnet options
 	if (gnunet_opt_y != NULL)
@@ -938,9 +827,16 @@ int main
 		    (arg.find("-e") == 0) || (arg.find("-x") == 0) ||
 			(arg.find("-P") == 0) || (arg.find("-H") == 0) ||
 		    (arg.find("-u") == 0) || (arg.find("-U") == 0) ||
-			(arg.find("-k") == 0) || (arg.find("-y") == 0))
+			(arg.find("-K") == 0) || (arg.find("-k") == 0) ||
+			(arg.find("-f") == 0) || (arg.find("-y") == 0))
 		{
 			size_t idx = ++i;
+			if ((arg.find("-f") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_fingerprint == NULL))
+			{
+				fingerprint = argv[i+1];
+				opt_fingerprint = (char*)fingerprint.c_str();
+			}
 			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) &&
 				(opt_ifilename == NULL))
 			{
@@ -977,6 +873,12 @@ int main
 				URI = argv[i+1];
 				opt_URI = (char*)URI.c_str();
 			}
+			if ((arg.find("-K") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_K == NULL))
+			{
+				Kfilename = argv[i+1];
+				opt_K = (char*)Kfilename.c_str();
+			}
 			if ((arg.find("-k") == 0) && (idx < (size_t)(argc - 1)) &&
 				(opt_k == NULL))
 			{
@@ -1006,10 +908,11 @@ int main
 			}
 			continue;
 		}
-		else if ((arg.find("--") == 0) || (arg.find("-r") == 0) ||
-			(arg.find("-v") == 0) || (arg.find("-h") == 0) ||
-			(arg.find("-V") == 0) || (arg.find("-1") == 0) ||
-			(arg.find("-2") == 0) || (arg.find("-3") == 0))
+		else if ((arg.find("--") == 0) || (arg.find("-a") == 0) ||
+			(arg.find("-r") == 0) || (arg.find("-v") == 0) ||
+			(arg.find("-h") == 0) || (arg.find("-V") == 0) ||
+			(arg.find("-1") == 0) || (arg.find("-2") == 0) ||
+			(arg.find("-3") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -1018,19 +921,25 @@ int main
 				std::cout << about << std::endl;
 				std::cout << "Arguments mandatory for long options are also" <<
 					" mandatory for short options." << std::endl;
-				std::cout << "  -h, --help       print this help" << std::endl;
 				std::cout << "  -1, --one        issuer has not done any" <<
 					" verification of claim of identity" << std::endl;
 				std::cout << "  -2, --two        issuer has done some casual" <<
 					" verification of claim of identity" << std::endl;
 				std::cout << "  -3, --three      issuer has done substantial" <<
 					" verification of claim of identity" << std::endl;
-				std::cout << "  -e TIME          expiration time of" <<
+				std::cout << "  -a, --ask        require confirmation from" <<
+					" STDIN for each signature" << std::endl;
+				std::cout << "  -e INTEGER       expiration time of" <<
 					" generated signatures in seconds" << std::endl;
+				std::cout << "  -f STRING        fingerprint of the public" <<
+					" key for certification" << std::endl;
 				std::cout << "  -H STRING        hostname (e.g. onion" <<
 					" address) of this peer within PEERS" << std::endl;
+				std::cout << "  -h, --help       print this help" << std::endl;
 				std::cout << "  -i FILENAME      create certification" <<
 					" signatures on key from FILENAME" << std::endl;
+				std::cout << "  -K FILENAME      select public key for" <<
+					" certification from keyring FILENAME" << std::endl;
 				std::cout << "  -k FILENAME      use keyring FILENAME" <<
 					" containing external revocation keys" << std::endl;
 				std::cout << "  -o FILENAME      write key with" <<
@@ -1050,13 +959,15 @@ int main
 					std::endl;
 				std::cout << "  -V, --verbose    turn on verbose output" <<
 					std::endl;
-				std::cout << "  -W TIME          timeout for point-to-point" <<
+				std::cout << "  -W INTEGER       timeout for point-to-point" <<
 					" messages in minutes" << std::endl;
 				std::cout << "  -y FILENAME    yet another OpenPGP tool with" <<
 					" private key in FILENAME" << std::endl;
 #endif
 				return 0; // not continue
 			}
+			if ((arg.find("-a") == 0) || (arg.find("--ask") == 0))
+				opt_a = true; // require confirmation
 			if ((arg.find("-r") == 0) || (arg.find("--revocation") == 0))
 				opt_r = true; // create revocation signature
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
@@ -1120,10 +1031,16 @@ int main
 #endif
 
 	// check command line arguments
-	if (opt_ifilename == NULL)
+	if ((opt_ifilename == NULL) && (opt_K == NULL))
 	{
-		std::cerr << "ERROR: option -i required to specify an input file" <<
-			std::endl;
+		std::cerr << "ERROR: option \"-i\" or \"-K\" is required to specify" <<
+			" an input file" << std::endl;
+		return -1;
+	}
+	if ((opt_K != NULL) && (opt_fingerprint == NULL))
+	{
+		std::cerr << "ERROR: option \"-f\" is required to select public key" <<
+			" for certification" << std::endl;
 		return -1;
 	}
 	if ((opt_hostname != NULL) && (opt_passwords == NULL) && (opt_y == NULL))
@@ -1138,12 +1055,7 @@ int main
 			usage << std::endl;
 		return -1;
 	}
-
-	// canonicalize peer list
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it =
-		std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
+	canonicalize(peers);
 	if (((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N)) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
@@ -1157,7 +1069,7 @@ int main
 	}
 
 	// lock memory
-	bool force_secmem = false;
+	bool force_secmem = false, should_unlock = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
@@ -1165,230 +1077,172 @@ int main
 		// at least try to use libgcrypt's secure memory
 		force_secmem = true;
 	}
+	else
+		should_unlock = true;
 
 	// initialize LibTMCG
 	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	if (opt_verbose)
 		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
 			std::endl;
 	
-	// initialize return code
+	// initialize return code and do the main work
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if ((opt_hostname != NULL) && (opt_y == NULL))
 	{
-		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
-		if ((opt_p < 1) || (opt_p > 65535))
-		{
-			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
-			return -1;
-		}
-		tcpip_init(hostname);
-		tcpip_bindports((uint16_t)opt_p, false);
-		tcpip_bindports((uint16_t)opt_p, true);
-		while (tcpip_connect((uint16_t)opt_p, false) < peers.size())
-			sleep(1);
-		while (tcpip_connect((uint16_t)opt_p, true) < peers.size())
-			sleep(1);
-		tcpip_accept();
-		tcpip_fork();
-		ret = tcpip_io();
-		tcpip_close();
-		tcpip_done();
-		return ret;
+		// start interactive variant, if built-in TCP/IP requested
+		ret = run_tcpip(peers.size(), opt_p, hostname, port);
 	}
 	else if (opt_y != NULL)
 	{
-		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
-		run_instance(0, time(NULL), opt_e, 0);
-		return ret;
+		// start a single instance as replacement for GnuPG et al.
+		fork_instance(0);
+		ret = wait_instance(0, opt_verbose, pid);
 	}
-
-	// start interactive variant with GNUnet or otherwise a local test
-#ifdef GNUNET
-	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
-		GNUNET_GETOPT_option_flag('1',
-			"one",
-			"issuer has not done any verification of the claim of identity",
-			&gnunet_opt_1
-		),
-		GNUNET_GETOPT_option_flag('2',
-			"two",
-			"issuer has done some casual verification of the claim of identity",
-			&gnunet_opt_2
-		),
-		GNUNET_GETOPT_option_flag('3',
-			"three",
-			"issuer has done substantial verification of the claim of identity",
-			&gnunet_opt_3
-		),
-		GNUNET_GETOPT_option_uint('e',
-			"expiration",
-			"TIME",
-			"expiration time of generated signature in seconds",
-			&gnunet_opt_sigexptime
-		),
-		GNUNET_GETOPT_option_string('H',
-			"hostname",
-			"STRING",
-			"hostname (e.g. onion address) of this peer within PEERS",
-			&gnunet_opt_hostname
-		),
-		GNUNET_GETOPT_option_string('i',
-			"input",
-			"FILENAME",
-			"create certification signature on key resp. user ID from FILENAME",
-			&gnunet_opt_ifilename
-		),
-		GNUNET_GETOPT_option_string('k',
-			"keyring",
-			"FILENAME",
-			"use keyring FILENAME containing external revocation keys",
-			&gnunet_opt_k
-		),
-		GNUNET_GETOPT_option_string('o',
-			"output",
-			"FILENAME",
-			"write key with certification signature attached to FILENAME",
-			&gnunet_opt_ofilename
-		),
-		GNUNET_GETOPT_option_string('p',
-			"port",
-			"STRING",
-			"GNUnet CADET port to listen/connect",
-			&gnunet_opt_port
-		),
-		GNUNET_GETOPT_option_string('P',
-			"passwords",
-			"STRING",
-			"exchanged passwords to protect private and broadcast channels",
-			&gnunet_opt_passwords
-		),
-		GNUNET_GETOPT_option_flag('r',
-			"revocation",
-			"create a certification revocation signature",
-			&gnunet_opt_r
-		),
-		GNUNET_GETOPT_option_string('u',
-			"userid",
-			"STRING",
-			"sign only valid user IDs containing STRING",
-			&gnunet_opt_u
-		),
-		GNUNET_GETOPT_option_string('U',
-			"URI",
-			"STRING",
-			"policy URI tied to signature",
-			&gnunet_opt_URI
-		),
-		GNUNET_GETOPT_option_flag('V',
-			"verbose",
-			"turn on verbose output",
-			&gnunet_opt_verbose
-		),
-		GNUNET_GETOPT_option_uint('w',
-			"wait",
-			"TIME",
-			"minutes to wait until start of signing protocol",
-			&gnunet_opt_wait
-		),
-		GNUNET_GETOPT_option_uint('W',
-			"aiou-timeout",
-			"TIME",
-			"timeout for point-to-point messages in minutes",
-			&gnunet_opt_W
-		),
-		GNUNET_GETOPT_option_string('y',
-			"yaot",
-			"FILNAME",
-			"yet another OpenPGP tool with private key in FILENAME",
-			&gnunet_opt_y
-		),
-		GNUNET_GETOPT_option_uint('x',
-			"x-tests",
-			NULL,
-			"number of exchange tests",
-			&gnunet_opt_xtests
-		),
-		GNUNET_GETOPT_OPTION_END
-	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
-		argv[0]);
-//	GNUNET_free((void *) argv);
-	if (ret == GNUNET_OK)
-		return 0;
 	else
-		return -1;
-#else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
-		" of this program" << std::endl;
-#endif
-
-	std::cerr << "INFO: running local test with " << peers.size() <<
-		" participants" << std::endl;
-	// open pipes
-	for (size_t i = 0; i < peers.size(); i++)
 	{
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if (pipe(pipefd[i][j]) < 0)
-				perror("ERROR: dkg-keysign (pipe)");
-			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("ERROR: dkg-keysign (pipe)");
-		}
-	}
-	
-	// start childs
-	for (size_t i = 0; i < peers.size(); i++)
-		fork_instance(i);
-
-	// sleep for five seconds
-	sleep(5);
-	
-	// wait for childs and close pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		int wstatus = 0;
-		if (opt_verbose)
-			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
-		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("ERROR: dkg-keysign (waitpid)");
-		if (!WIFEXITED(wstatus))
-		{
-			std::cerr << "ERROR: protocol instance ";
-			if (WIFSIGNALED(wstatus))
-				std::cerr << pid[i] << " terminated by signal " <<
-					WTERMSIG(wstatus) << std::endl;
-			if (WCOREDUMP(wstatus))
-				std::cerr << pid[i] << " dumped core" << std::endl;
+		// start interactive variant with GNUnet or otherwise a local test
+#ifdef GNUNET
+		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_flag('1',
+				"one",
+				"issuer has not done any verification of the claim of identity",
+				&gnunet_opt_1
+			),
+			GNUNET_GETOPT_option_flag('2',
+				"two",
+				"issuer has done some casual verification of the claim of identity",
+				&gnunet_opt_2
+			),
+			GNUNET_GETOPT_option_flag('3',
+				"three",
+				"issuer has done substantial verification of the claim of identity",
+				&gnunet_opt_3
+			),
+			GNUNET_GETOPT_option_flag('a',
+				"ask",
+				"require confirmation from STDIN for each signature",
+				&gnunet_opt_a
+			),
+			GNUNET_GETOPT_option_uint('e',
+				"expiration",
+				"INTEGER",
+				"expiration time of generated signature in seconds",
+				&gnunet_opt_sigexptime
+			),
+			GNUNET_GETOPT_option_string('f',
+				"fingerprint",
+				"STRING",
+				"fingerprint of the public key for certification",
+				&gnunet_opt_fingerprint
+			),
+			GNUNET_GETOPT_option_string('H',
+				"hostname",
+				"STRING",
+				"hostname (e.g. onion address) of this peer within PEERS",
+				&gnunet_opt_hostname
+			),
+			GNUNET_GETOPT_option_string('i',
+				"input",
+				"FILENAME",
+				"create certification signature on key resp. user ID from FILENAME",
+				&gnunet_opt_ifilename
+			),
+			GNUNET_GETOPT_option_string('K',
+				"keys",
+				"FILENAME",
+				"select public key for certification from keyring FILENAME",
+				&gnunet_opt_K
+			),
+			GNUNET_GETOPT_option_string('k',
+				"keyring",
+				"FILENAME",
+				"use keyring FILENAME containing external revocation keys",
+				&gnunet_opt_k
+			),
+			GNUNET_GETOPT_option_string('o',
+				"output",
+				"FILENAME",
+				"write key with certification signature attached to FILENAME",
+				&gnunet_opt_ofilename
+			),
+			GNUNET_GETOPT_option_string('p',
+				"port",
+				"STRING",
+				"GNUnet CADET port to listen/connect",
+				&gnunet_opt_port
+			),
+			GNUNET_GETOPT_option_string('P',
+				"passwords",
+				"STRING",
+				"exchanged passwords to protect private and broadcast channels",
+				&gnunet_opt_passwords
+			),
+			GNUNET_GETOPT_option_flag('r',
+				"revocation",
+				"create a certification revocation signature",
+				&gnunet_opt_r
+			),
+			GNUNET_GETOPT_option_string('u',
+				"userid",
+				"STRING",
+				"sign only valid user IDs containing STRING",
+				&gnunet_opt_u
+			),
+			GNUNET_GETOPT_option_string('U',
+				"URI",
+				"STRING",
+				"policy URI tied to signature",
+				&gnunet_opt_URI
+			),
+			GNUNET_GETOPT_option_flag('V',
+				"verbose",
+				"turn on verbose output",
+				&gnunet_opt_verbose
+			),
+			GNUNET_GETOPT_option_uint('w',
+				"wait",
+				"INTEGER",
+				"minutes to wait until start of signing protocol",
+				&gnunet_opt_wait
+			),
+			GNUNET_GETOPT_option_uint('W',
+				"aiou-timeout",
+				"INTEGER",
+				"timeout for point-to-point messages in minutes",
+				&gnunet_opt_W
+			),
+			GNUNET_GETOPT_option_string('y',
+				"yaot",
+				"FILNAME",
+				"yet another OpenPGP tool with private key in FILENAME",
+				&gnunet_opt_y
+			),
+			GNUNET_GETOPT_option_uint('x',
+				"x-tests",
+				NULL,
+				"number of exchange tests",
+				&gnunet_opt_xtests
+			),
+			GNUNET_GETOPT_OPTION_END
+		};
+		ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions,
+			&gnunet_run, argv[0]);
+		if (ret != GNUNET_OK)
 			ret = -1;
-		}
-		else if (WIFEXITED(wstatus))
-		{
-			if (opt_verbose)
-				std::cerr << "INFO: protocol instance " << pid[i] <<
-					" terminated with exit status " << WEXITSTATUS(wstatus) <<
-					std::endl;
-			if (WEXITSTATUS(wstatus))
-				ret = -2; // error
-		}
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-keysign (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
-				(close(broadcast_pipefd[i][j][1]) < 0))
-			{
-				perror("ERROR: dkg-keysign (close)");
-			}
-		}
+#else
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd, self_pipefd,
+			broadcast_pipefd, broadcast_self_pipefd, &fork_instance);
+#endif
 	}
-	
+	if (should_unlock)
+		unlock_memory();
 	return ret;
 }
 
