@@ -57,7 +57,9 @@ static const char *protocol = "DKGPG-sign-1.0";
 #include "dkg-common.hh"
 
 int 						pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							self_pipefd[2];
 int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_self_pipefd[2];
 pid_t 						pid[DKGPG_MAX_N];
 std::vector<std::string>	peers;
 bool						instance_forked = false;
@@ -76,6 +78,7 @@ char							*opt_k = NULL;
 char							*opt_y = NULL;
 unsigned long int				opt_e = 0, opt_p = 55000, opt_W = 5;
 bool							opt_t = false, opt_E = false, opt_C = false;
+bool							opt_v5 = false;
 
 void run_instance
 	(size_t whoami, const time_t sigtime, const time_t sigexptime,
@@ -236,10 +239,16 @@ void run_instance
 				// simple key -- we assume that GNUnet provides secure channels
 				key << "dkg-sign::p_" << (i + whoami);
 			}
-			uP_in.push_back(pipefd[i][whoami][0]);
+			if (i == whoami)
+				uP_in.push_back(self_pipefd[0]);
+			else
+				uP_in.push_back(pipefd[i][whoami][0]);
 			uP_out.push_back(pipefd[whoami][i][1]);
 			uP_key.push_back(key.str());
-			bP_in.push_back(broadcast_pipefd[i][whoami][0]);
+			if (i == whoami)
+				bP_in.push_back(broadcast_self_pipefd[0]);
+			else
+				bP_in.push_back(broadcast_pipefd[i][whoami][0]);
 			bP_out.push_back(broadcast_pipefd[whoami][i][1]);
 			bP_key.push_back(key.str());
 		}
@@ -282,76 +291,81 @@ void run_instance
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // fixed hash algo SHA2-512
 	}
 
-	// compute the hash of the input file
+	// prepare signature and compute the hash of the input file
 	if (opt_verbose)
 		std::cerr << "INFO: hashing the input file \"" << opt_ifilename <<
 			"\"" << std::endl;
 	tmcg_openpgp_octets_t trailer, hash, left;
+	bool hret = false;
 	if (opt_t || opt_C)
 	{
-		if (opt_y == NULL)
+		if (opt_v5)
+		{
+			CallasDonnerhackeFinneyShawThayerRFC4880::
+				PacketSigPrepareDetachedSignatureV5(
+					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT,
+					prv->pub->pkalgo, hashalgo, csigtime, sigexptime, URI,
+					prv->pub->fingerprint, trailer);
+			tmcg_openpgp_octets_t htrailer(trailer);				
+			// For detached signatures 6 zero bytes are hashed instead.
+			for (size_t i = 0; i < 6; i++)
+				htrailer.push_back(0);
+			hret = CallasDonnerhackeFinneyShawThayerRFC4880::
+				TextDocumentHashV5(opt_ifilename, htrailer, hashalgo, hash,
+				left);
+		}
+		else
 		{
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareDetachedSignature(
 					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT,
-					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
-					trailer);
-		}
-		else
-		{
-			CallasDonnerhackeFinneyShawThayerRFC4880::
-				PacketSigPrepareDetachedSignature(
-					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT, prv->pkalgo,
-					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
-					trailer);
-		}
-		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
-			TextDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
-		{
-			std::cerr << "ERROR: p_" << whoami << ": TextDocumentHash()" <<
-				" failed; cannot process input file \"" << opt_ifilename <<
-				"\"" << std::endl;
-			if (opt_y == NULL)
-			{
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-			}
-			delete prv;
-			exit(-1);
+					prv->pub->pkalgo, hashalgo, csigtime, sigexptime, URI,
+					prv->pub->fingerprint, trailer);
+			hret = CallasDonnerhackeFinneyShawThayerRFC4880::
+				TextDocumentHash(opt_ifilename, trailer, hashalgo, hash, left);
 		}
 	}
 	else
 	{
-		if (opt_y == NULL)
+		if (opt_v5)
 		{
 			CallasDonnerhackeFinneyShawThayerRFC4880::
-				PacketSigPrepareDetachedSignature(
-					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT,
+				PacketSigPrepareDetachedSignatureV5(
+					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT, prv->pub->pkalgo,
 					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
 					trailer);
+			tmcg_openpgp_octets_t htrailer(trailer);				
+			// For detached signatures 6 zero bytes are hashed instead.
+			for (size_t i = 0; i < 6; i++)
+				htrailer.push_back(0);
+			hret = CallasDonnerhackeFinneyShawThayerRFC4880::
+				BinaryDocumentHashV5(opt_ifilename, htrailer, hashalgo, hash,
+				left);
 		}
 		else
 		{
 			CallasDonnerhackeFinneyShawThayerRFC4880::
 				PacketSigPrepareDetachedSignature(
-					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT, prv->pkalgo,
+					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT, prv->pub->pkalgo,
 					hashalgo, csigtime, sigexptime, URI, prv->pub->fingerprint,
 					trailer);
+			hret = CallasDonnerhackeFinneyShawThayerRFC4880::
+				BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash,
+					left);
 		}
-		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
-			BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
+	}
+	if (!hret)
+	{
+		std::cerr << "ERROR: p_" << whoami << ": [Text|Binary]DocumentHash" <<
+			"[V5]() failed; cannot process input file \"" << opt_ifilename <<
+			"\"" << std::endl;
+		if (opt_y == NULL)
 		{
-			std::cerr << "ERROR: p_" << whoami << ": BinaryDocumentHash()" <<
-				" failed; cannot process input file \"" << opt_ifilename <<
-				"\"" << std::endl;
-			if (opt_y == NULL)
-			{
-				delete rbc, delete aiou, delete aiou2;
-				delete dss;
-			}
-			delete prv;
-			exit(-1);
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
 		}
+		delete prv;
+		exit(-1);
 	}
 
 	// sign the hash value
@@ -380,16 +394,12 @@ void run_instance
 		rbc->Sync(synctime);
 		// release RBC
 		delete rbc;
-		// release handles (unicast channel)
+		// release handles (both channels)
 		if (opt_verbose)
 		{
 			std::cerr << "INFO: p_" << whoami << ": unicast channels";
 			aiou->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
-		}
-		// release handles (broadcast channel)
-		if (opt_verbose)
-		{
 			std::cerr << "INFO: p_" << whoami << ": broadcast channel";
 			aiou2->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
@@ -450,6 +460,7 @@ int gnunet_opt_verbose = 0;
 int gnunet_opt_t = 0;
 int gnunet_opt_E = 0;
 int gnunet_opt_C = 0;
+int gnunet_opt_v5 = 0;
 #endif
 
 void fork_instance
@@ -462,6 +473,21 @@ void fork_instance
 		if (pid[whoami] == 0)
 		{
 			/* BEGIN child code: participant p_i */
+			if ((self_pipefd[0] == -1) && (self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				self_pipefd[0] = dup(pipefd[whoami][whoami][0]);
+				self_pipefd[1] = dup(pipefd[whoami][whoami][1]);
+			}
+			if ((broadcast_self_pipefd[0] == -1) &&
+				(broadcast_self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				broadcast_self_pipefd[0] =
+					dup(broadcast_pipefd[whoami][whoami][0]);
+				broadcast_self_pipefd[1] =
+					dup(broadcast_pipefd[whoami][whoami][1]);
+			}
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_sigexptime,
@@ -493,6 +519,11 @@ int main
 	char *cfg_fn = NULL;
 	static const struct GNUNET_GETOPT_CommandLineOption options[] = {
 		GNUNET_GETOPT_option_cfgfile(&cfg_fn),
+		GNUNET_GETOPT_option_flag('5',
+			"v5",
+			"generate a v5 signature (cf. RFC 4880bis-06)",
+			&gnunet_opt_v5
+		),
 		GNUNET_GETOPT_option_flag('C',
 			"clear",
 			"apply cleartext signature framework (cf. RFC 4880)",
@@ -703,7 +734,7 @@ int main
 		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) ||
 			(arg.find("-h") == 0) || (arg.find("-V") == 0) ||
 			(arg.find("-t") == 0) || (arg.find("-E") == 0) ||
-			(arg.find("-C") == 0))
+			(arg.find("-C") == 0) || (arg.find("-5") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -712,6 +743,8 @@ int main
 				std::cout << about << std::endl;
 				std::cout << "Arguments mandatory for long options are also" <<
 					" mandatory for short options." << std::endl;
+				std::cout << "  -5, --v5       generate a v5 signature" <<
+					std::endl;
 				std::cout << "  -C, --clear    apply cleartext signature" <<
 					" framework (cf. RFC 4880)" << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
@@ -752,6 +785,8 @@ int main
 				opt_E = true;
 			if ((arg.find("-t") == 0) || (arg.find("--text") == 0))
 				opt_t = true;
+			if ((arg.find("-5") == 0) || (arg.find("--v5") == 0))
+				opt_v5 = true;
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
 			{
 #ifndef GNUNET
@@ -879,6 +914,11 @@ int main
 		// start interactive variant with GNUnet or otherwise a local test
 #ifdef GNUNET
 		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_flag('5',
+				"v5",
+				"generate a v5 signature (cf. RFC 4880bis-06)",
+				&gnunet_opt_v5
+			),
 			GNUNET_GETOPT_option_flag('C',
 				"clear",
 				"apply cleartext signature framework (cf. RFC 4880)",
@@ -978,8 +1018,8 @@ int main
 		if (ret != GNUNET_OK)
 			ret = -1;
 #else
-		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
-			broadcast_pipefd, &fork_instance);
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd, self_pipefd,
+			broadcast_pipefd, broadcast_self_pipefd, &fork_instance);
 #endif
 	}
 	if (should_unlock)
@@ -987,4 +1027,3 @@ int main
 	return ret;
 }
 
-
