@@ -56,11 +56,13 @@ static const char *protocol = "DKGPG-addrevoker-1.0";
 #include "dkg-io.hh"
 #include "dkg-common.hh"
 
-int								pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-int								broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-pid_t							pid[DKGPG_MAX_N];
-std::vector<std::string>		peers;
-bool							instance_forked = false;
+int							pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							self_pipefd[2];
+int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_self_pipefd[2];
+pid_t						pid[DKGPG_MAX_N];
+std::vector<std::string>	peers;
+bool						instance_forked = false;
 
 tmcg_openpgp_secure_string_t	passphrase;
 std::string						ifilename, kfilename;
@@ -273,10 +275,16 @@ void run_instance
 				// simple key -- we assume that GNUnet provides secure channels
 				key << "dkg-addrevoker::p_" << (i + whoami);
 			}
-			uP_in.push_back(pipefd[i][whoami][0]);
+			if (i == whoami)
+				uP_in.push_back(self_pipefd[0]);
+			else
+				uP_in.push_back(pipefd[i][whoami][0]);
 			uP_out.push_back(pipefd[whoami][i][1]);
 			uP_key.push_back(key.str());
-			bP_in.push_back(broadcast_pipefd[i][whoami][0]);
+			if (i == whoami)
+				bP_in.push_back(broadcast_self_pipefd[0]);
+			else
+				bP_in.push_back(broadcast_pipefd[i][whoami][0]);
 			bP_out.push_back(broadcast_pipefd[whoami][i][1]);
 			bP_key.push_back(key.str());
 		}
@@ -321,14 +329,11 @@ void run_instance
 	}
 
 	// prepare trailer and hash for the direct-key signature
-	tmcg_openpgp_octets_t trailer, issuer, revoker, hash, left;
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(prv->pub->pub_hashing, issuer);
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(primary->pub_hashing, revoker);
+	tmcg_openpgp_octets_t trailer, hash, left;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		PacketSigPrepareDesignatedRevoker(prv->pub->pkalgo, hashalgo, csigtime,
-			prv->pub->flags, issuer, primary->pkalgo, revoker, trailer);
+			prv->pub->flags, prv->pub->fingerprint, primary->pkalgo,
+			primary->fingerprint, trailer);
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		KeyHash(prv->pub->pub_hashing, trailer, hashalgo, hash, left);
 	delete primary;
@@ -361,16 +366,12 @@ void run_instance
 		rbc->Sync(synctime);
 		// release RBC
 		delete rbc;
-		// release handles (unicast channel)
+		// release handles (both channels)
 		if (opt_verbose)
 		{
 			std::cerr << "INFO: p_" << whoami << ": unicast channels";
 			aiou->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
-		}
-		// release handles (broadcast channel)
-		if (opt_verbose)
-		{
 			std::cerr << "INFO: p_" << whoami << ": broadcast channel";
 			aiou2->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
@@ -452,6 +453,21 @@ void fork_instance
 		if (pid[whoami] == 0)
 		{
 			/* BEGIN child code: participant p_i */
+			if ((self_pipefd[0] == -1) && (self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				self_pipefd[0] = dup(pipefd[whoami][whoami][0]);
+				self_pipefd[1] = dup(pipefd[whoami][whoami][1]);
+			}
+			if ((broadcast_self_pipefd[0] == -1) &&
+				(broadcast_self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				broadcast_self_pipefd[0] =
+					dup(broadcast_pipefd[whoami][whoami][0]);
+				broadcast_self_pipefd[1] =
+					dup(broadcast_pipefd[whoami][whoami][1]);
+			}
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_xtests);
@@ -851,8 +867,8 @@ int main
 		if (ret != GNUNET_OK)
 			ret = -1;
 #else
-		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
-			broadcast_pipefd, &fork_instance);
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd, self_pipefd,
+			broadcast_pipefd, broadcast_self_pipefd, &fork_instance);
 #endif
 	}
 	if (should_unlock)
