@@ -309,19 +309,19 @@ void xtest
 {
 	for (size_t i = 0; i < num_xtests; i++)
 	{
-		mpz_t xtest;
-		mpz_init_set_ui(xtest, i);
-		std::cerr << "INFO: p_" << whoami << ": xtest = " << xtest << " <-> ";
-		rbc->Broadcast(xtest);
+		mpz_t x;
+		mpz_init_set_ui(x, i);
+		std::cerr << "INFO: p_" << whoami << ": x = " << x << " <-> ";
+		rbc->Broadcast(x);
 		for (size_t ii = 0; ii < peers; ii++)
 		{
-			if (!rbc->DeliverFrom(xtest, ii))
+			if (!rbc->DeliverFrom(x, ii))
 				std::cerr << "<X> ";
 			else
-				std::cerr << xtest << " ";
+				std::cerr << x << " ";
 		}
 		std::cerr << std::endl;
-		mpz_clear(xtest);
+		mpz_clear(x);
 	}
 }
 
@@ -341,19 +341,17 @@ time_t agree_time
 	tvs.push_back(mytime);
 	for (size_t i = 0; i < peers; i++)
 	{
-		if (i != whoami)
+		if (rbc->DeliverFrom(mtv, i))
 		{
-			if (rbc->DeliverFrom(mtv, i))
-			{
-				time_t utv;
-				utv = (time_t)mpz_get_ui(mtv);
+			time_t utv;
+			utv = (time_t)mpz_get_ui(mtv);
+			if (i != whoami)
 				tvs.push_back(utv);
-			}
-			else
-			{
-				std::cerr << "WARNING: p_" << whoami << ": no creation" <<
-					" timestamp received from p_" << i << std::endl;
-			}
+		}
+		else
+		{
+			std::cerr << "WARNING: p_" << whoami << ": no creation" <<
+				" timestamp received from p_" << i << std::endl;
 		}
 	}
 	mpz_clear(mtv);
@@ -605,7 +603,9 @@ int run_localtest
 	 const int opt_verbose,
 	 pid_t pid[DKGPG_MAX_N],
 	 int pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2],
+	 int self_pipefd[2],
 	 int bpipefd[DKGPG_MAX_N][DKGPG_MAX_N][2],
+	 int bself_pipefd[2],
 	 void (*fork_instance)(const size_t))
 {
 	assert(peers <= DKGPG_MAX_N);
@@ -623,7 +623,11 @@ int run_localtest
 				perror("ERROR: dkg-common:run_localtest (pipe)");
 		}
 	}
-	
+
+	// initialize self-pipes
+	self_pipefd[0] = -1, self_pipefd[1] = -1;
+	bself_pipefd[0] = -1, bself_pipefd[1] = -1;
+
 	// start childs
 	for (size_t i = 0; i < peers; i++)
 		fork_instance(i);
@@ -640,9 +644,9 @@ int run_localtest
 		for (size_t j = 0; j < peers; j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-common:run_localtest (close)");
+				perror("WARNING: dkg-common:run_localtest (close)");
 			if ((close(bpipefd[i][j][0]) < 0) || (close(bpipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-common:run_localtest (close)");
+				perror("WARNING: dkg-common:run_localtest (close)");
 		}
 	}
 	return ret;
