@@ -35,7 +35,7 @@
 // copy infos from DKGPG package before overwritten by GNUnet headers
 static const char *version = PACKAGE_VERSION " (" PACKAGE_NAME ")";
 static const char *about = PACKAGE_STRING " " PACKAGE_URL;
-static const char *protocol = "DKGPG-generate-1.0";
+static const char *protocol = "DKGPG-generate-1.1";
 
 #include <sstream>
 #include <fstream>
@@ -57,7 +57,9 @@ static const char *protocol = "DKGPG-generate-1.0";
 #include "dkg-common.hh"
 
 int							pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							self_pipefd[2];
 int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_self_pipefd[2];
 pid_t						pid[DKGPG_MAX_N];
 std::vector<std::string>	peers;
 bool						instance_forked = false;
@@ -132,10 +134,11 @@ void run_instance
 		tmcg_mpz_ssrandomm_cache(cache, cache_mod, cache_avail, dsa_x, vtmf->q);
 		tmcg_mpz_spowm(dsa_y, vtmf->g, dsa_x, vtmf->p);
 		// extract parameters for OpenPGP key structures
-		std::string out, crcout, armor;
-		tmcg_openpgp_octets_t all, pub, sec, uid, uidsig;
+		std::string armor;
+		tmcg_openpgp_octets_t all, pub, sec, dirsig, uid, uidsig;
 		tmcg_openpgp_octets_t sub, ssb, subsig, dsaflags, elgflags, issuer;
 		tmcg_openpgp_octets_t pub_hashing, sub_hashing;
+		tmcg_openpgp_octets_t dirsig_hashing, dirsig_left;
 		tmcg_openpgp_octets_t uidsig_hashing, uidsig_left;
 		tmcg_openpgp_octets_t subsig_hashing, subsig_left;
 		tmcg_openpgp_octets_t hash, empty;
@@ -232,6 +235,35 @@ void run_instance
 			dsaflags.push_back(0x08); // "This key may be used for timestamping"
 		}
 		sigtime = time(NULL); // current time
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			PacketSigPrepareDesignatedRevoker(TMCG_OPENPGP_PKALGO_DSA, hashalgo,
+				sigtime, dsaflags, issuer, (tmcg_openpgp_pkalgo_t)0, empty,
+				dirsig_hashing);
+		hash.clear();
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			KeyHash(pub_hashing, dirsig_hashing, hashalgo, hash, dirsig_left);
+		r = gcry_mpi_new(2048);
+		s = gcry_mpi_new(2048);
+		ret = CallasDonnerhackeFinneyShawThayerRFC4880::
+			AsymmetricSignDSA(hash, key, r, s);
+		if (ret)
+		{
+			std::cerr << "ERROR: AsymmetricSignDSA() failed" << std::endl;
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_mpi_release(r);
+			gcry_mpi_release(s);
+			gcry_sexp_release(key);
+			delete vtmf;
+			exit(-1);
+		}
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			PacketSigEncode(dirsig_hashing, dirsig_left, r, s, dirsig);
+		gcry_mpi_release(r);
+		gcry_mpi_release(s);
 		CallasDonnerhackeFinneyShawThayerRFC4880::
 			PacketSigPrepareSelfSignature(TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION,
 				hashalgo, sigtime, keyexptime, dsaflags, issuer, uidsig_hashing); 
@@ -347,6 +379,7 @@ void run_instance
 		pubfilename << peers[whoami] << "-pub.asc";
 		armor = "", all.clear();
 		all.insert(all.end(), pub.begin(), pub.end());
+		all.insert(all.end(), dirsig.begin(), dirsig.end());
 		all.insert(all.end(), uid.begin(), uid.end());
 		all.insert(all.end(), uidsig.begin(), uidsig.end());
 		all.insert(all.end(), sub.begin(), sub.end());
@@ -373,6 +406,7 @@ void run_instance
 		secfilename << peers[whoami] << "-sec.asc";
 		armor = "", all.clear();
 		all.insert(all.end(), sec.begin(), sec.end());
+		all.insert(all.end(), dirsig.begin(), dirsig.end());
 		all.insert(all.end(), uid.begin(), uid.end());
 		all.insert(all.end(), uidsig.begin(), uidsig.end());
 		all.insert(all.end(), ssb.begin(), ssb.end());
@@ -452,10 +486,16 @@ void run_instance
 			// use simple key -- we assume that GNUnet provides secure channels
 			key << "dkg-generate::P_" << (i + whoami);
 		}
-		uP_in.push_back(pipefd[i][whoami][0]);
+		if (i == whoami)
+			uP_in.push_back(self_pipefd[0]);
+		else
+			uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
 		uP_key.push_back(key.str());
-		bP_in.push_back(broadcast_pipefd[i][whoami][0]);
+		if (i == whoami)
+			bP_in.push_back(broadcast_self_pipefd[0]);
+		else
+			bP_in.push_back(broadcast_pipefd[i][whoami][0]);
 		bP_out.push_back(broadcast_pipefd[whoami][i][1]);
 		bP_key.push_back(key.str());
 	}
@@ -692,10 +732,11 @@ void run_instance
 		tmcg_mpz_spowm(dsa_y, vtmf->g, dsa_x, vtmf->p);
 	}
 	// extract further public parameters for OpenPGP key structures
-	std::string out, crcout, armor;
-	tmcg_openpgp_octets_t all, pub, sec, uid, uidsig;
+	std::string armor;
+	tmcg_openpgp_octets_t all, pub, sec, dirsig, uid, uidsig;
 	tmcg_openpgp_octets_t sub, ssb, subsig, dsaflags, elgflags, issuer;
 	tmcg_openpgp_octets_t pub_hashing, sub_hashing;
+	tmcg_openpgp_octets_t dirsig_hashing, dirsig_left;
 	tmcg_openpgp_octets_t uidsig_hashing, uidsig_left;
 	tmcg_openpgp_octets_t subsig_hashing, subsig_left;
 	tmcg_openpgp_octets_t hash, empty;
@@ -966,7 +1007,8 @@ void run_instance
 	}
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		PacketBodyExtract(pub, 0, pub_hashing);
-	CallasDonnerhackeFinneyShawThayerRFC4880::FingerprintCompute(pub_hashing, issuer);
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		FingerprintCompute(pub_hashing, issuer);
 	CallasDonnerhackeFinneyShawThayerRFC4880::PacketUidEncode(userid, uid);
 	// "In a V4 key, the primary key MUST be a key capable of certification."
 	if (S > 0)
@@ -994,11 +1036,151 @@ void run_instance
 		// for a non-shared DSA primary key no common timestamp required 
 		sigtime = time(NULL); // current time
 	}
-	// create an additional (!) direct-key signature (0x1f) with above key flags
-// TODO: "The split key (0x10) [...] flags are placed on a self-signature only; they
-//        are meaningless on a certification signature. They SHOULD be placed only
-//        on a direct-key signature (type 0x1f) or a subkey signature (type 0x18),
-//        one that refers to the key the flag applies to." [RFC 4880]
+	// create an additional direct-key signature (0x1f) with above key flags
+	// "The split key (0x10) [...] flags are placed on a self-signature only;
+	//  they are meaningless on a certification signature. They SHOULD be
+	//  placed only on a direct-key signature (type 0x1f) or a subkey signature
+	//  (type 0x18), one that refers to the key the flag applies to." [RFC 4880]
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		PacketSigPrepareDesignatedRevoker(TMCG_OPENPGP_PKALGO_DSA, hashalgo,
+			sigtime, dsaflags, issuer, (tmcg_openpgp_pkalgo_t)0, empty,
+			dirsig_hashing);
+	hash.clear();
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		KeyHash(pub_hashing, dirsig_hashing, hashalgo, hash, dirsig_left);
+	if (S > 0)
+	{
+		tmcg_openpgp_byte_t buffer[1024];
+		gcry_mpi_t h;
+		size_t buflen = 0;
+		memset(buffer, 0, sizeof(buffer));
+		for (size_t i = 0; i < hash.size(); i++, buflen++)
+		{
+			if (i < sizeof(buffer))
+				buffer[i] = hash[i];
+		}
+		ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
+		if (ret)
+		{
+			std::cerr << "ERROR: P_" << whoami <<
+				": gcry_mpi_scan() failed for h" << std::endl;
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_sexp_release(key);
+			delete dkg, delete dss;
+			delete rbc, delete vtmf, delete aiou, delete aiou2;
+			exit(-1);
+		}
+		if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
+		{
+			std::cerr << "ERROR: P_" << whoami <<
+				": tmcg_mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
+			gcry_mpi_release(h);
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_sexp_release(key);
+			delete dkg, delete dss;
+			delete rbc, delete vtmf, delete aiou, delete aiou2;
+			exit(-1);
+		}
+		gcry_mpi_release(h);
+		std::stringstream err_log_sign;
+		if (opt_verbose)
+			std::cerr << "INFO: P_" << whoami <<
+				": dss.Sign() for direct-key signature (0x1f)" << std::endl;
+		if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, aiou, rbc,
+			err_log_sign))
+		{
+			std::cerr << "ERROR: P_" << whoami << ": tDSS Sign() failed" <<
+				std::endl;
+			std::cerr << "ERROR: P_" << whoami << ": log follows " <<
+				std::endl << err_log_sign.str();
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_sexp_release(key);
+			delete dkg, delete dss;
+			delete rbc, delete vtmf, delete aiou, delete aiou2;
+			exit(-1);
+		}
+		if (opt_verbose > 1)
+			std::cerr << "INFO: P_" << whoami << ": log follows " <<
+				std::endl << err_log_sign.str();
+		r = gcry_mpi_new(2048);
+		if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
+		{
+			std::cerr << "ERROR: P_" << whoami <<
+				": tmcg_mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_mpi_release(r);
+			gcry_sexp_release(key);
+			delete dkg, delete dss;
+			delete rbc, delete vtmf, delete aiou, delete aiou2;
+			exit(-1);
+		}
+		s = gcry_mpi_new(2048);
+		if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
+		{
+			std::cerr << "ERROR: P_" << whoami <<
+				": tmcg_mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_mpi_release(r);
+			gcry_mpi_release(s);
+			gcry_sexp_release(key);
+			delete dkg, delete dss;
+			delete rbc, delete vtmf, delete aiou, delete aiou2;
+			exit(-1);
+		}
+	}
+	else
+	{
+		r = gcry_mpi_new(2048);
+		s = gcry_mpi_new(2048);
+		ret = CallasDonnerhackeFinneyShawThayerRFC4880::
+			AsymmetricSignDSA(hash, key, r, s);
+		if (ret)
+		{
+			std::cerr << "ERROR: P_" << whoami <<
+				": AsymmetricSignDSA() failed" << std::endl;
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			gcry_mpi_release(p);
+			gcry_mpi_release(q);
+			gcry_mpi_release(g);
+			gcry_mpi_release(y);
+			gcry_mpi_release(x);
+			gcry_mpi_release(r);
+			gcry_mpi_release(s);
+			gcry_sexp_release(key);
+			delete dkg, delete dss;
+			delete rbc, delete vtmf, delete aiou, delete aiou2;
+			exit(-1);
+		}
+	}
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		PacketSigEncode(dirsig_hashing, dirsig_left, r, s, dirsig);
+	gcry_mpi_release(r);
+	gcry_mpi_release(s);
 	// create a positive certification (0x13) of the included user ID
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		PacketSigPrepareSelfSignature(TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION,
@@ -1362,19 +1544,19 @@ void run_instance
 		if (S > 0)
 		{
 			tmcg_openpgp_byte_t buffer[1024];
-			gcry_mpi_t h;
+			gcry_mpi_t ha;
 			size_t buflen = 0;
 			memset(buffer, 0, sizeof(buffer));
-			for (size_t i = 0; i < hash.size(); i++, buflen++)
+			for (size_t j = 0; j < hash.size(); j++, buflen++)
 			{
-				if (i < sizeof(buffer))
-					buffer[i] = hash[i];
+				if (j < sizeof(buffer))
+					buffer[j] = hash[j];
 			}
-			ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
+			ret = gcry_mpi_scan(&ha, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 			if (ret)
 			{
 				std::cerr << "ERROR: P_" << whoami <<
-					": gcry_mpi_scan() failed for h" << std::endl;
+					": gcry_mpi_scan() failed for ha" << std::endl;
 				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 				gcry_mpi_release(p);
 				gcry_mpi_release(q);
@@ -1384,11 +1566,11 @@ void run_instance
 				delete rbc, delete vtmf, delete aiou, delete aiou2;
 				exit(-1);
 			}
-			if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
+			if (!tmcg_mpz_set_gcry_mpi(ha, dsa_m))
 			{
 				std::cerr << "ERROR: P_" << whoami <<
 					": tmcg_mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
-				gcry_mpi_release(h);
+				gcry_mpi_release(ha);
 				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 				gcry_mpi_release(p);
 				gcry_mpi_release(q);
@@ -1398,7 +1580,7 @@ void run_instance
 				delete rbc, delete vtmf, delete aiou, delete aiou2;
 				exit(-1);
 			}
-			gcry_mpi_release(h);
+			gcry_mpi_release(ha);
 			std::stringstream err_log_sign;
 			if (opt_verbose)
 				std::cerr << "INFO: P_" << whoami <<
@@ -1533,6 +1715,7 @@ void run_instance
 	pubfilename << peers[whoami] << "_dkg-pub.asc";
 	armor = "", all.clear();
 	all.insert(all.end(), pub.begin(), pub.end());
+	all.insert(all.end(), dirsig.begin(), dirsig.end());
 	all.insert(all.end(), uid.begin(), uid.end());
 	all.insert(all.end(), uidsig.begin(), uidsig.end());
 	all.insert(all.end(), sub.begin(), sub.end());
@@ -1550,6 +1733,7 @@ void run_instance
 	std::string sfilename = secfilename.str();
 	armor = "", all.clear();
 	all.insert(all.end(), sec.begin(), sec.end());
+	all.insert(all.end(), dirsig.begin(), dirsig.end());
 	all.insert(all.end(), uid.begin(), uid.end());
 	all.insert(all.end(), uidsig.begin(), uidsig.end());
 	all.insert(all.end(), ssb.begin(), ssb.end());
@@ -2001,12 +2185,27 @@ void fork_instance
 	(const size_t whoami)
 {
 	if ((pid[whoami] = fork()) < 0)
-		perror("ERROR: dkg-generate (fork)");
+		perror("ERROR: dkg-generate:fork_instance (fork)");
 	else
 	{
 		if (pid[whoami] == 0)
 		{
 			/* BEGIN child code: participant P_i */
+			if ((self_pipefd[0] == -1) && (self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				self_pipefd[0] = dup(pipefd[whoami][whoami][0]);
+				self_pipefd[1] = dup(pipefd[whoami][whoami][1]);
+			}
+			if ((broadcast_self_pipefd[0] == -1) &&
+				(broadcast_self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				broadcast_self_pipefd[0] =
+					dup(broadcast_pipefd[whoami][whoami][0]);
+				broadcast_self_pipefd[1] =
+					dup(broadcast_pipefd[whoami][whoami][1]);
+			}
 			time_t keytime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, keytime, gnunet_opt_keyexptime,
@@ -2346,7 +2545,7 @@ int main
 	peers.push_back("Test4");
 	opt_verbose = 2;
 	opt_e = 10800;
-	if (tmcg_mpz_wrandom_ui() % 2)
+	if (tmcg_mpz_wrandom_ui() % 2) // FIXME: libgcrypt is not initialized yet
 	{
 		// sometimes test a non-FIPS CRS
 		crs = "crs|VMyMoPc2vb51ofxb4f2rebOSONnfhitfGcYxdav2D4wqBTeZrC"
@@ -2386,7 +2585,7 @@ int main
 	opt_verbose = 2;
 	opt_e = 10800;
 	opt_timestamping = true;
-	if (tmcg_mpz_wrandom_ui() % 2)
+	if (tmcg_mpz_wrandom_ui() % 2) // FIXME: libgcrypt is not initialized yet
 	{
 		// sometimes test a non-FIPS CRS
 		crs = "crs|VMyMoPc2vb51ofxb4f2rebOSONnfhitfGcYxdav2D4wqBTeZrC"
@@ -2706,8 +2905,8 @@ int main
 		if (ret != GNUNET_OK)
 			ret = -1;
 #else
-		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
-			broadcast_pipefd, &fork_instance);
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd, self_pipefd,
+			broadcast_pipefd, broadcast_self_pipefd, &fork_instance);
 #endif
 	}
 	// release cache
