@@ -34,7 +34,9 @@
 #include "dkg-gnunet-common.hh"
 
 extern int				pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+extern int				self_pipefd[2];
 extern int				broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+extern int				broadcast_self_pipefd[2];
 extern pid_t			pid[DKGPG_MAX_N];
 extern std::vector<std::string>		peers;
 extern bool				instance_forked;
@@ -150,8 +152,10 @@ void handle_gnunet_data_callback
 	GNUNET_assert(ntohs(message->size) >= sizeof(*message));
 	uint16_t len = ntohs(message->size) - sizeof(*message);
 	if ((gnunet_opt_verbose) && (opt_verbose > 1))
+	{
 		GNUNET_log(GNUNET_ERROR_TYPE_MESSAGE, "Got message of type %u from %s"
 			" with %u bytes\n", ntohs(message->type), peer.c_str(), len);
+	}
 	const char *buf = (const char *)&message[1];
 	int fd;
 	if (ntohs(message->type) == GNUNET_MESSAGE_TYPE_TMCG_DKG_PIPE_UNICAST)
@@ -205,37 +209,69 @@ void gnunet_pipe_ready
 	pt = NULL;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
-		if (i != peer2pipe[thispeer])
+		char *th_buf = new char[pipe_buffer_size];
+		ssize_t num = read(pipefd[peer2pipe[thispeer]][i][0], th_buf,
+			pipe_buffer_size);
+		if (num < 0)
 		{
-			char *th_buf = new char[pipe_buffer_size];
-			ssize_t num = read(pipefd[peer2pipe[thispeer]][i][0], th_buf,
-				pipe_buffer_size);
-			if (num < 0)
+			delete [] th_buf;
+			if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
 			{
-				delete [] th_buf;
-				if ((errno == EAGAIN) || (errno == EWOULDBLOCK) ||
-					(errno == EINTR))
-				{
-					continue;
-				}
-				else
-				{
-					perror("ERROR: dkg-gnunet-common (read)");
-					GNUNET_SCHEDULER_shutdown();
-					return;
-				}
-			}
-			else if (num == 0)
-			{
-				delete [] th_buf;
 				continue;
 			}
 			else
+			{
+				perror("ERROR: dkg-gnunet-common (read)");
+				GNUNET_SCHEDULER_shutdown();
+				return;
+			}
+		}
+		else if (num == 0)
+		{
+			delete [] th_buf;
+			continue;
+		}
+		else
+		{
+			if (i != peer2pipe[thispeer])
 			{
 				DKG_BufferListEntry ble = DKG_BufferListEntry(i,
 					DKG_Buffer(num, th_buf));
 				send_queue.push_back(ble);
 			}
+			else
+			{
+				ssize_t wlen = 0;
+				do
+				{
+					ssize_t wnum = write(self_pipefd[1],
+						th_buf + wlen, num - wlen);
+					if (wnum < 0)
+					{
+						if ((errno == EWOULDBLOCK) || (errno == EINTR))
+						{
+							sleep(1);
+							continue;
+						}
+						else if (errno == EAGAIN)
+						{
+							perror("WARNING: dkg-gnunet-common (write)");
+							sleep(1);
+							continue;
+						}
+						else
+						{
+							perror("ERROR: dkg-gnunet-common (write)");
+							GNUNET_SCHEDULER_shutdown();
+							return;
+						}
+					}
+					else
+						wlen += wnum;
+				}
+				while (wlen < num);
+				delete [] th_buf;
+			}
 		}
 	}
 	// reschedule I/O task
@@ -251,37 +287,69 @@ void gnunet_broadcast_pipe_ready
 	pt_broadcast = NULL;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
-		if (i != peer2pipe[thispeer])
+		char *th_buf = new char[pipe_buffer_size];
+		ssize_t num = read(broadcast_pipefd[peer2pipe[thispeer]][i][0],
+			th_buf, pipe_buffer_size);
+		if (num < 0)
 		{
-			char *th_buf = new char[pipe_buffer_size];
-			ssize_t num = read(broadcast_pipefd[peer2pipe[thispeer]][i][0],
-				th_buf, pipe_buffer_size);
-			if (num < 0)
+			delete [] th_buf;
+			if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
 			{
-				delete [] th_buf;
-				if ((errno == EAGAIN) || (errno == EWOULDBLOCK) ||
-					(errno == EINTR))
-				{
-					continue;
-				}
-				else
-				{
-					perror("ERROR: dkg-gnunet-common (read)");
-					GNUNET_SCHEDULER_shutdown();
-					return;
-				}
-			}
-			else if (num == 0)
-			{
-				delete [] th_buf;
 				continue;
 			}
 			else
+			{
+				perror("ERROR: dkg-gnunet-common (read)");
+				GNUNET_SCHEDULER_shutdown();
+				return;
+			}
+		}
+		else if (num == 0)
+		{
+			delete [] th_buf;
+			continue;
+		}
+		else
+		{
+			if (i != peer2pipe[thispeer])
 			{
 				DKG_BufferListEntry ble = DKG_BufferListEntry(i,
 					DKG_Buffer(num, th_buf));
 				send_queue_broadcast.push_back(ble);
 			}
+			else
+			{
+				ssize_t wlen = 0;
+				do
+				{
+					ssize_t wnum = write(broadcast_self_pipefd[1],
+						th_buf + wlen, num - wlen);
+					if (wnum < 0)
+					{
+						if ((errno == EWOULDBLOCK) || (errno == EINTR))
+						{
+							sleep(1);
+							continue;
+						}
+						else if (errno == EAGAIN)
+						{
+							perror("WARNING: dkg-gnunet-common (write)");
+							sleep(1);
+							continue;
+						}
+						else
+						{
+							perror("ERROR: dkg-gnunet-common (write)");
+							GNUNET_SCHEDULER_shutdown();
+							return;
+						}
+					}
+					else
+						wlen += wnum;
+				}
+				while (wlen < num);
+				delete [] th_buf;
+			}
 		}
 	}
 	// reschedule I/O task
@@ -300,16 +368,20 @@ void gnunet_channel_ended
 		if (pipe2channel_out.count(i) && (pipe2channel_out[i] == channel))
 		{
 			if (gnunet_opt_verbose)
+			{
 				std::cerr << "WARNING: output channel ended for peer = " <<
 					pipe2peer[i] << std::endl;
+			}
 			pipe2channel_out.erase(i);
 			return;
 		}
 		if (pipe2channel_in.count(i) && (pipe2channel_in[i] == channel))
 		{
 			if (gnunet_opt_verbose)
+			{
 				std::cerr << "WARNING: input channel ended for peer = " <<
 					pipe2peer[i] << std::endl;
+			}
 			pipe2channel_in.erase(i);
 			return;
 		}
@@ -324,8 +396,10 @@ void* gnunet_channel_incoming
 	if (cls == NULL)
 		cls = NULL; // dummy code to supress compiler warning
 	if (gnunet_opt_verbose)
+	{
 		std::cerr << "INFO: incoming channel from " <<
 			GNUNET_i2s_full(initiator) << std::endl;
+	}
 	// check whether peer identity is included in peer list
 	std::string peer = GNUNET_i2s_full(initiator);
 	if (peer2pipe.count(peer) == 0)
@@ -485,9 +559,11 @@ void gnunet_io
 			struct GNUNET_MessageHeader *msg;
 			// send message on input channel
 			if ((gnunet_opt_verbose) && (opt_verbose > 1))
+			{
 				std::cerr << "INFO: try to send " << buf.first <<
 					" bytes on input channel to " << pipe2peer[ble.first] <<
 					std::endl;
+			}
 			env = GNUNET_MQ_msg_extra(msg, buf.first,
 				GNUNET_MESSAGE_TYPE_TMCG_DKG_PIPE_UNICAST);
 			GNUNET_memcpy(&msg[1], buf.second, buf.first);
@@ -509,9 +585,11 @@ void gnunet_io
 			struct GNUNET_MessageHeader *msg;
 			// send message on input channel
 			if ((gnunet_opt_verbose) && (opt_verbose > 1))
+			{
 				std::cerr << "INFO: try to broadcast " << buf.first <<
 					" bytes on input channel to " << pipe2peer[ble.first] <<
 					std::endl;
+			}
 			env = GNUNET_MQ_msg_extra(msg, buf.first,
 				GNUNET_MESSAGE_TYPE_TMCG_DKG_PIPE_BROADCAST);
 			GNUNET_memcpy(&msg[1], buf.second, buf.first);
@@ -561,11 +639,11 @@ void gnunet_connect
 			if (pipe2channel_out.count(i))
 				GNUNET_CADET_channel_destroy(pipe2channel_out[i]);
 			// create new CADET output channels
-			struct GNUNET_PeerIdentity pid;
+			struct GNUNET_PeerIdentity pi;
 			enum GNUNET_CADET_ChannelOption flags = GNUNET_CADET_OPTION_RELIABLE;
 			struct GNUNET_CADET_Channel *ch;
 			if (GNUNET_CRYPTO_eddsa_public_key_from_string(pipe2peer[i].c_str(),
-				pipe2peer[i].length(), &pid.public_key) != GNUNET_OK)
+				pipe2peer[i].length(), &pi.public_key) != GNUNET_OK)
 			{
 				std::cerr << "ERROR: bad public key of peer = " <<
 					pipe2peer[i] << std::endl;
@@ -590,7 +668,7 @@ void gnunet_connect
 					NULL),
 				GNUNET_MQ_handler_end()
 			};
-			ch = GNUNET_CADET_channel_create(mh, (void*)i, &pid, &porthash,
+			ch = GNUNET_CADET_channel_create(mh, (void*)i, &pi, &porthash,
 				flags, NULL, &gnunet_channel_ended, handlers);
 			if (ch == NULL)
 			{
@@ -618,13 +696,14 @@ void gnunet_statistics
 	st = NULL;
 
 	if ((gnunet_opt_verbose) && (opt_verbose > 1))
+	{
 		std::cerr << "INFO: pipe2channel_out.size() = " <<
 			pipe2channel_out.size() << ", pipe2channel_in.size() = " <<
 			pipe2channel_in.size() << std::endl;
-	if ((gnunet_opt_verbose) && (opt_verbose > 1))
 		std::cerr << "INFO: send_queue.size() = " << send_queue.size() <<
 			", send_queue_broadcast.size() = " << send_queue_broadcast.size() <<
 			std::endl;
+	}
 	if (instance_forked)
 	{
 		// shutdown, if forked instance has terminated 
@@ -639,17 +718,21 @@ void gnunet_statistics
 			{
 				std::cerr << "ERROR: protocol instance ";
 				if (WIFSIGNALED(wstatus))
+				{
 					std::cerr << thispid << " terminated by signal " <<
 						WTERMSIG(wstatus) << std::endl;
+				}
 				if (WCOREDUMP(wstatus))
 					std::cerr << thispid << " dumped core" << std::endl;
 			}
 			else if (WIFEXITED(wstatus))
 			{
 				if (gnunet_opt_verbose)
+				{
 					std::cerr << "INFO: protocol instance " << thispid <<
 						" terminated with exit status " <<
 						WEXITSTATUS(wstatus) << std::endl;
+				}
 			}
 			instance_forked = false;
 			GNUNET_SCHEDULER_shutdown();
@@ -722,6 +805,8 @@ void gnunet_init
 				perror("ERROR: dkg-gnunet-common (pipe)");
 		}
 	}
+	self_pipefd[0] = -1, self_pipefd[1] = -1;
+	broadcast_self_pipefd[0] = -1, broadcast_self_pipefd[1] = -1;
 	pipes_created = true;
 
 	// next: schedule connect, fork and statistics tasks
@@ -791,8 +876,10 @@ void gnunet_run(void *cls, char *const *args, const char *cfgfile,
 	else
 		GNUNET_CRYPTO_hash(port.c_str(), port.length(), &porthash);
 	if (gnunet_opt_verbose)
+	{
 		std::cerr << "INFO: my CADET listen port hash = " <<
 			GNUNET_h2s_full(&porthash) << std::endl;
+	}
 	static const struct GNUNET_MQ_MessageHandler handlers[] = {
 		GNUNET_MQ_handler_end()
 	};
