@@ -53,7 +53,9 @@ static const char *protocol = "DKGPG-revoke-1.0";
 #include "dkg-common.hh"
 
 int 						pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							self_pipefd[2];
 int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_self_pipefd[2];
 pid_t 						pid[DKGPG_MAX_N];
 std::vector<std::string>	peers;
 bool						instance_forked = false;
@@ -208,10 +210,16 @@ void run_instance
 			// use simple key -- we assume that GNUnet provides secure channels
 			key << "dkg-revoke::p_" << (i + whoami);
 		}
-		uP_in.push_back(pipefd[i][whoami][0]);
+		if (i == whoami)
+			uP_in.push_back(self_pipefd[0]);
+		else
+			uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
 		uP_key.push_back(key.str());
-		bP_in.push_back(broadcast_pipefd[i][whoami][0]);
+		if (i == whoami)
+			bP_in.push_back(broadcast_self_pipefd[0]);
+		else
+			bP_in.push_back(broadcast_pipefd[i][whoami][0]);
 		bP_out.push_back(broadcast_pipefd[whoami][i][1]);
 		bP_key.push_back(key.str());
 	}
@@ -464,6 +472,21 @@ void fork_instance
 		if (pid[whoami] == 0)
 		{
 			/* BEGIN child code: participant p_i */
+			if ((self_pipefd[0] == -1) && (self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				self_pipefd[0] = dup(pipefd[whoami][whoami][0]);
+				self_pipefd[1] = dup(pipefd[whoami][whoami][1]);
+			}
+			if ((broadcast_self_pipefd[0] == -1) &&
+				(broadcast_self_pipefd[1] == -1))
+			{
+				// duplication of file descriptors for run_localtest()
+				broadcast_self_pipefd[0] =
+					dup(broadcast_pipefd[whoami][whoami][0]);
+				broadcast_self_pipefd[1] =
+					dup(broadcast_pipefd[whoami][whoami][1]);
+			}
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_r, gnunet_opt_xtests);
@@ -573,10 +596,10 @@ int main
 		opt_hostname = gnunet_opt_hostname;
 	if (gnunet_opt_passwords != NULL)
 		opt_passwords = gnunet_opt_passwords;
-	if (gnunet_opt_passwords != NULL)
-		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
+	if (gnunet_opt_passwords != NULL)
+		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_k != NULL)
 		opt_k = gnunet_opt_k;
 	if (gnunet_opt_R != NULL)
@@ -841,8 +864,8 @@ int main
 		if (ret != GNUNET_OK)
 			ret = -1;
 #else
-	ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
-		broadcast_pipefd, &fork_instance);
+	ret = run_localtest(peers.size(), opt_verbose, pid, pipefd, self_pipefd,
+		broadcast_pipefd, broadcast_self_pipefd, &fork_instance);
 #endif
 	}
 	if (should_unlock)
