@@ -25,7 +25,9 @@
 #include "dkg-tcpip-common.hh"
 
 extern int						pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+extern int						self_pipefd[2];
 extern int						broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+extern int						broadcast_self_pipefd[2];
 extern pid_t					pid[DKGPG_MAX_N];
 extern std::vector<std::string>	peers;
 extern bool						instance_forked;
@@ -54,14 +56,18 @@ RETSIGTYPE tcpip_sig_handler_quit(int sig)
 	if (instance_forked && (pid[tcpip_peer2pipe[tcpip_thispeer]] == 0))
 	{
 		if (opt_verbose)
+		{
 			std::cerr << "tcpip_sig_handler_quit(): child got signal " <<
 				sig << std::endl;
+		}
 	}
 	else
 	{
 		if (opt_verbose)
+		{
 			std::cerr << "tcpip_sig_handler_quit(): parent got signal " <<
 				sig << std::endl;
+		}
 		// remove our own signal handler and quit
 		signal(SIGINT, SIG_DFL);
 		signal(SIGQUIT, SIG_DFL);
@@ -79,8 +85,10 @@ void tcpip_init
 	tcpip_thispeer = hostname;
 	// initialize peer2pipe and pipe2peer mapping
 	if (opt_verbose)
+	{
 		std::cerr << "INFO: using built-in TCP/IP service for message" <<
 			" exchange instead of GNUnet CADET" << std::endl;
+	}
 	if (std::find(peers.begin(), peers.end(), hostname) == peers.end())
 	{
 		std::cerr << "ERROR: cannot find hostname \"" << hostname << "\" of" <<
@@ -99,16 +107,26 @@ void tcpip_init
 		{
 			if (pipe2(pipefd[i][j], O_NONBLOCK) < 0)
 			{
-				perror("ERROR: dkg-tcpip-common (pipe)");
+				perror("ERROR: dkg-tcpip-common (pipe2)");
 				exit(-1);
 			}
 			if (pipe2(broadcast_pipefd[i][j], O_NONBLOCK) < 0)
 			{
-				perror("ERROR: dkg-tcpip-common (pipe)");
+				perror("ERROR: dkg-tcpip-common (pipe2)");
 				exit(-1);
 			}
 		}
 	}
+	if (pipe2(self_pipefd, O_NONBLOCK) < 0)
+	{
+		perror("ERROR: dots-tcpip-common (pipe2)");
+		exit(-1);
+	}
+	if (pipe2(broadcast_self_pipefd, O_NONBLOCK) < 0)
+	{
+		perror("ERROR: dots-tcpip-common (pipe2)");
+		exit(-1);
+	}	
 	// install our own signal handler
 	struct sigaction act;
 	memset(&act, 0, sizeof(act));
@@ -134,10 +152,11 @@ void tcpip_bindports
 	(const uint16_t stpo, const bool broadcast)
 {
 	if (opt_verbose > 2)
+	{
 		std::cerr << "INFO: tcpip_bindports(" << stpo << ", " <<
 			(broadcast ? "true" : "false") << ") called" << std::endl;
-	if (opt_verbose > 2)
 		std::cerr << "INFO: FD_SETSIZE = " << FD_SETSIZE << std::endl;
+	}
 	uint16_t peers_size = 0;
 	if (peers.size() <= DKGPG_MAX_N)
 	{
@@ -174,7 +193,7 @@ void tcpip_bindports
 		{
 			std::cerr << "ERROR: resolving wildcard address failed: ";
 			if (ret == EAI_SYSTEM)
-				perror("ERROR: dkg-tcpip-common (getaddrinfo)");
+				perror("dkg-tcpip-common (getaddrinfo)");
 			else
 				std::cerr << gai_strerror(ret);
 			std::cerr << std::endl;
@@ -191,6 +210,12 @@ void tcpip_bindports
 				perror("WARNING: dkg-tcpip-common (socket)");
 				continue; // try next address
 			}
+			int sockopt = 1;
+			if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &sockopt,
+				sizeof(sockopt)) < 0)
+			{
+				perror("WARNING: dkg-tcpip-common (setsockopt)");
+			}
 			char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
 			memset(hbuf, 0, sizeof(hbuf));
 			memset(sbuf, 0, sizeof(sbuf));
@@ -200,25 +225,35 @@ void tcpip_bindports
 			{
 				std::cerr << "ERROR: resolving wildcard address failed: ";
 				if (ret == EAI_SYSTEM)
-					perror("ERROR: dkg-tcpip-common (getnameinfo)");
+					perror("dkg-tcpip-common (getnameinfo)");
 				else
 					std::cerr << gai_strerror(ret);
 				std::cerr << std::endl;
 				if (close(sockfd) < 0)
-					perror("ERROR: dkg-tcpip-common (close)");
+					perror("WARNING: dkg-tcpip-common (close)");
 				freeaddrinfo(res);
 				tcpip_close();
 				tcpip_done();
 				exit(-1);
 			}
 			if (opt_verbose)
+			{
 				std::cerr << "INFO: bind TCP/IP port " << port <<
 					" at address " << hbuf << std::endl;
-			if (bind(sockfd, rp->ai_addr, rp->ai_addrlen) < 0)
+			}
+			time_t btime = time(NULL), bto = DKGPG_TCPIP_BIND_TIMEOUT;
+			int bret;
+			do
+			{
+				sleep(1);
+				bret = bind(sockfd, rp->ai_addr, rp->ai_addrlen);
+			}
+			while ((bret == EADDRINUSE) && (time(NULL) < (btime + bto))); 
+			if (bret < 0)
 			{
 				perror("WARNING: dkg-tcpip-common (bind)");
 				if (close(sockfd) < 0)
-					perror("ERROR: dkg-tcpip-common (close)");
+					perror("WARNING: dkg-tcpip-common (close)");
 				sockfd = -1;
 				continue; // try next address
 			}
@@ -237,7 +272,7 @@ void tcpip_bindports
 		{
 			perror("ERROR: dkg-tcpip-common (listen)");
 			if (close(sockfd) < 0)
-				perror("ERROR: dkg-tcpip-common (close)");
+				perror("WARNING: dkg-tcpip-common (close)");
 			tcpip_close();
 			tcpip_done();
 			exit(-1);
@@ -253,8 +288,10 @@ size_t tcpip_connect
 	(const uint16_t stpo, const bool broadcast)
 {
 	if (opt_verbose > 2)
+	{
 		std::cerr << "INFO: tcpip_connect(" << stpo << ", " <<
 			(broadcast ? "true" : "false") << ") called" << std::endl;
+	}
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		if ((broadcast && (tcpip_broadcast_pipe2socket_out.count(i) == 0)) ||
@@ -289,7 +326,7 @@ size_t tcpip_connect
 				std::cerr << "ERROR: resolving hostname \"" <<
 					peers[i] << "\" failed: ";
 				if (ret == EAI_SYSTEM)
-					perror("ERROR: dkg-tcpip-common (getaddrinfo)");
+					perror("dkg-tcpip-common (getaddrinfo)");
 				else
 					std::cerr << gai_strerror(ret);
 				std::cerr << std::endl;
@@ -311,7 +348,7 @@ size_t tcpip_connect
 					if (errno != ECONNREFUSED)
 						perror("WARNING: dkg-tcpip-common (connect)");					
 					if (close(sockfd) < 0)
-						perror("ERROR: dkg-tcpip-common (close)");
+						perror("WARNING: dkg-tcpip-common (close)");
 					continue; // try next address
 				}
 				else
@@ -326,23 +363,24 @@ size_t tcpip_connect
 						std::cerr << "ERROR: resolving hostname \"" <<
 							peers[i] << "\" failed: ";
 						if (ret == EAI_SYSTEM)
-							perror("ERROR: dkg-tcpip-common (getnameinfo)");
+							perror("dkg-tcpip-common (getnameinfo)");
 						else
 							std::cerr << gai_strerror(ret);
 						std::cerr << std::endl;
 						if (close(sockfd) < 0)
-							perror("ERROR: dkg-tcpip-common (close)");
+							perror("WARNING: dkg-tcpip-common (close)");
 						freeaddrinfo(res);
 						tcpip_close();
 						tcpip_done();
 						exit(-1);
 					}
 					if (opt_verbose)
+					{
 						std::cerr << "INFO: resolved hostname \"" <<
 							peers[i] << "\" to address " << hbuf << std::endl;
-					if (opt_verbose)
 						std::cerr << "INFO: connected to host \"" <<
 							peers[i] << "\" on port " << port << std::endl;
+					}
 					if (broadcast)
 						tcpip_broadcast_pipe2socket_out[i] = sockfd;
 					else
@@ -363,9 +401,10 @@ void tcpip_accept
 	()
 {
 	if (opt_verbose > 2)
+	{
 		std::cerr << "INFO: tcpip_accept(...) called" << std::endl;
-	if (opt_verbose > 2)
 		std::cerr << "INFO: FD_SETSIZE = " << FD_SETSIZE << std::endl;
+	}
 	while ((tcpip_pipe2socket_in.size() < peers.size()) ||
 		(tcpip_broadcast_pipe2socket_in.size() < peers.size()))
 	{
@@ -454,7 +493,7 @@ void tcpip_accept
 				{
 					std::cerr << "ERROR: resolving incoming address failed: ";
 					if (ret == EAI_SYSTEM)
-						perror("ERROR: dkg-tcpip-common (getnameinfo)");
+						perror("dkg-tcpip-common (getnameinfo)");
 					else
 						std::cerr << gai_strerror(ret);
 					std::cerr << std::endl;
@@ -463,8 +502,10 @@ void tcpip_accept
 					exit(-1);
 				}
 				if (opt_verbose)
+				{
 					std::cerr << "INFO: accept connection for p_" <<
 						pi->first << " from address " << ipaddr << std::endl;
+				}
 			}
 		}
 		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket.begin();
@@ -489,7 +530,7 @@ void tcpip_accept
 				{
 					std::cerr << "ERROR: resolving incoming address failed: ";
 					if (ret == EAI_SYSTEM)
-						perror("ERROR: dkg-tcpip-common (getnameinfo)");
+						perror("dkg-tcpip-common (getnameinfo)");
 					else
 						std::cerr << gai_strerror(ret);
 					std::cerr << std::endl;
@@ -498,9 +539,11 @@ void tcpip_accept
 					exit(-1);
 				}
 				if (opt_verbose)
+				{
 					std::cerr << "INFO: accept broadcast connection for" <<
 						" p_" << pi->first << " from address " << 
 						ipaddr << std::endl;
+				}
 			}
 		}
 	}
@@ -539,7 +582,7 @@ int tcpip_io
 			int thispid = pid[thisidx];
 			int ret = waitpid(thispid, &wstatus, WNOHANG);
 			if (ret < 0)
-				perror("ERROR: dkg-tcpip-common (waitpid)");
+				perror("WARNING: dkg-tcpip-common (waitpid)");
 			else if (ret == thispid)
 			{
 				instance_forked = false;
@@ -547,8 +590,10 @@ int tcpip_io
 				{
 					std::cerr << "ERROR: protocol instance ";
 					if (WIFSIGNALED(wstatus))
+					{
 						std::cerr << thispid << " terminated by signal " <<
 							WTERMSIG(wstatus) << std::endl;
+					}
 					if (WCOREDUMP(wstatus))
 						std::cerr << thispid << " dumped core" << std::endl;
 					return -1;
@@ -556,9 +601,11 @@ int tcpip_io
 				else if (WIFEXITED(wstatus))
 				{
 					if (opt_verbose)
+					{
 						std::cerr << "INFO: protocol instance " << thispid <<
 							" terminated with exit status " <<
 							WEXITSTATUS(wstatus) << std::endl;
+					}
 					return WEXITSTATUS(wstatus);
 				}
 				return 0;
@@ -571,77 +618,68 @@ int tcpip_io
 		for (tcpip_mci_t pi = tcpip_pipe2socket_in.begin();
 			pi != tcpip_pipe2socket_in.end(); ++pi)
 		{
-			if (pi->first != thisidx)
+			if (pi->second < FD_SETSIZE)
 			{
-				if (pi->second < FD_SETSIZE)
-				{
-					FD_SET(pi->second, &rfds);
-					if (pi->second > maxfd)
-						maxfd = pi->second;
-				}
-				else
-				{
-					std::cerr << "ERROR: file descriptor value of internal" <<
-						" pipe exceeds FD_SETSIZE" << std::endl;
-					tcpip_close();
-					tcpip_done();
-					exit(-1);
-				}
+				FD_SET(pi->second, &rfds);
+				if (pi->second > maxfd)
+					maxfd = pi->second;
+			}
+			else
+			{
+				std::cerr << "ERROR: file descriptor value of internal" <<
+					" pipe exceeds FD_SETSIZE" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
 			}
 		}
 		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
 			pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
 		{
-			if (pi->first != thisidx)
+			if (pi->second < FD_SETSIZE)
 			{
-				if (pi->second < FD_SETSIZE)
-				{
-					FD_SET(pi->second, &rfds);
-					if (pi->second > maxfd)
-						maxfd = pi->second;
-				}
-				else
-				{
-					std::cerr << "ERROR: file descriptor value of internal" <<
-						" pipe exceeds FD_SETSIZE" << std::endl;
-					tcpip_close();
-					tcpip_done();
-					exit(-1);
-				}
+				FD_SET(pi->second, &rfds);
+				if (pi->second > maxfd)
+					maxfd = pi->second;
+			}
+			else
+			{
+				std::cerr << "ERROR: file descriptor value of internal" <<
+					" pipe exceeds FD_SETSIZE" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
 			}
 		}
 		for (size_t i = 0; i < peers.size(); i++)
 		{
-			if (i != thisidx)
+			if (pipefd[thisidx][i][0] < FD_SETSIZE)
 			{
-				if (pipefd[thisidx][i][0] < FD_SETSIZE)
-				{
-					FD_SET(pipefd[thisidx][i][0], &rfds);
-					if (pipefd[thisidx][i][0] > maxfd)
-						maxfd = pipefd[thisidx][i][0];
-				}
-				else
-				{
-					std::cerr << "ERROR: file descriptor value of internal" <<
-						" pipe exceeds FD_SETSIZE" << std::endl;
-					tcpip_close();
-					tcpip_done();
-					exit(-1);
-				}
-				if (broadcast_pipefd[thisidx][i][0] < FD_SETSIZE)
-				{
-					FD_SET(broadcast_pipefd[thisidx][i][0], &rfds);
-					if (broadcast_pipefd[thisidx][i][0] > maxfd)
-						maxfd = broadcast_pipefd[thisidx][i][0];
-				}
-				else
-				{
-					std::cerr << "ERROR: file descriptor value of internal" <<
-						" pipe exceeds FD_SETSIZE" << std::endl;
-					tcpip_close();
-					tcpip_done();
-					exit(-1);
-				}
+				FD_SET(pipefd[thisidx][i][0], &rfds);
+				if (pipefd[thisidx][i][0] > maxfd)
+					maxfd = pipefd[thisidx][i][0];
+			}
+			else
+			{
+				std::cerr << "ERROR: file descriptor value of internal" <<
+					" pipe exceeds FD_SETSIZE" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
+			}
+			if (broadcast_pipefd[thisidx][i][0] < FD_SETSIZE)
+			{
+				FD_SET(broadcast_pipefd[thisidx][i][0], &rfds);
+				if (broadcast_pipefd[thisidx][i][0] > maxfd)
+					maxfd = broadcast_pipefd[thisidx][i][0];
+			}
+			else
+			{
+				std::cerr << "ERROR: file descriptor value of internal" <<
+					" pipe exceeds FD_SETSIZE" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
 			}
 		}
 		tv.tv_sec = 1;
@@ -668,7 +706,7 @@ int tcpip_io
 		for (tcpip_mci_t pi = tcpip_pipe2socket_in.begin();
 			pi != tcpip_pipe2socket_in.end(); ++pi)
 		{
-			if ((pi->first != thisidx) && FD_ISSET(pi->second, &rfds))
+			if (FD_ISSET(pi->second, &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
 				ssize_t len = read(pi->second, buf, sizeof(buf));
@@ -702,20 +740,33 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
+					{
 						std::cerr << "INFO: received " << len << " bytes on " <<
 							"connection for p_" << pi->first << std::endl;
+					}
 					ssize_t wnum = 0;
 					do
 					{
-						ssize_t num = write(pipefd[pi->first][thisidx][1],
+						ssize_t num = 0;
+						if (pi->first == thisidx)
+						{
+							num = write(self_pipefd[1],
 								buf + wnum, len - wnum);
+						}
+						else
+						{
+							num = write(pipefd[pi->first][thisidx][1],
+								buf + wnum, len - wnum);
+						}
 						if (num < 0)
 						{
 							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into pipe ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -723,8 +774,10 @@ int tcpip_io
 							{
 								perror("WARNING: dkg-tcpip-common (write)");
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into pipe ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -746,7 +799,7 @@ int tcpip_io
 		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
 			pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
 		{
-			if ((pi->first != thisidx) && FD_ISSET(pi->second, &rfds))
+			if (FD_ISSET(pi->second, &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
 				ssize_t len = read(pi->second, buf, sizeof(buf));
@@ -780,22 +833,34 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
+					{
 						std::cerr << "INFO: received " << len << " bytes on" <<
 							" broadcast connection for p_" <<
 							pi->first << std::endl;
+					}
 					ssize_t wnum = 0;
 					do
 					{
-						ssize_t num = write(
-							broadcast_pipefd[pi->first][thisidx][1], buf + wnum,
-							len - wnum);
+						ssize_t num = 0;
+						if (pi->first == thisidx)
+						{
+							num = write(broadcast_self_pipefd[1],
+								buf + wnum, len - wnum);
+						}
+						else
+						{
+							num = write(broadcast_pipefd[pi->first][thisidx][1],
+								buf + wnum, len - wnum);
+						}
 						if (num < 0)
 						{
 							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into pipe ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -803,8 +868,10 @@ int tcpip_io
 							{
 								perror("WARNING: dkg-tcpip-common (write)");
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into pipe ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -825,7 +892,7 @@ int tcpip_io
 		}
 		for (size_t i = 0; i < peers.size(); i++)
 		{
-			if ((i != thisidx) && FD_ISSET(pipefd[thisidx][i][0], &rfds))
+			if (FD_ISSET(pipefd[thisidx][i][0], &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
 				ssize_t len = read(pipefd[thisidx][i][0], buf, sizeof(buf));
@@ -856,8 +923,10 @@ int tcpip_io
 				else if (tcpip_pipe2socket_out.count(i))
 				{
 					if (opt_verbose > 1)
+					{
 						std::cerr << "INFO: sending " << len << " bytes on" <<
 							" connection to p_" << i << std::endl;
+					}
 					ssize_t wnum = 0;
 					do
 					{
@@ -868,8 +937,10 @@ int tcpip_io
 							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into socket ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -877,8 +948,10 @@ int tcpip_io
 							{
 								perror("WARNING: dkg-tcpip-common (write)");
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into socket ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -906,12 +979,13 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
+					{
 						std::cerr << "INFO: discarding " << len << " bytes" <<
 							" for p_" << i << std::endl;
+					}
 				}
 			}
-			if ((i != thisidx) &&
-				FD_ISSET(broadcast_pipefd[thisidx][i][0], &rfds))
+			if (FD_ISSET(broadcast_pipefd[thisidx][i][0], &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
 				ssize_t len = read(broadcast_pipefd[thisidx][i][0], buf,
@@ -942,9 +1016,10 @@ int tcpip_io
 				else if (tcpip_broadcast_pipe2socket_out.count(i))
 				{
 					if (opt_verbose > 1)
+					{
 						std::cerr << "INFO: sending " << len << " bytes on" <<
-							" broadcast connection to p_" << i <<
-							std::endl;
+							" broadcast connection to p_" << i << std::endl;
+					}
 					ssize_t wnum = 0;
 					do
 					{
@@ -955,8 +1030,10 @@ int tcpip_io
 							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into socket ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -964,8 +1041,10 @@ int tcpip_io
 							{
 								perror("WARNING: dkg-tcpip-common (write)");
 								if (opt_verbose)
+								{
 									std::cerr << "INFO: sleeping for write" <<
 										" into socket ..." << std::endl;
+								}
 								sleep(1);
 								continue;
 							}
@@ -993,8 +1072,10 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
+					{
 						std::cerr << "INFO: discarding " << len << " bytes" <<
 							" for p_" << i << std::endl;
+					}
 				}
 			}
 		}
@@ -1008,37 +1089,37 @@ void tcpip_close
 		pi != tcpip_pipe2socket_in.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("ERROR: dkg-tcpip-common (close)");
+			perror("WARNING: dkg-tcpip-common (close)");
 	}
 	for (tcpip_mci_t pi = tcpip_pipe2socket_out.begin();
 		pi != tcpip_pipe2socket_out.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("ERROR: dkg-tcpip-common (close)");
+			perror("WARNING: dkg-tcpip-common (close)");
 	}
 	for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
 		pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("ERROR: dkg-tcpip-common (close)");
+			perror("WARNING: dkg-tcpip-common (close)");
 	}
 	for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_out.begin();
 		pi != tcpip_broadcast_pipe2socket_out.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("ERROR: dkg-tcpip-common (close)");
+			perror("WARNING: dkg-tcpip-common (close)");
 	}
 	for (tcpip_mci_t pi = tcpip_pipe2socket.begin();
 		pi != tcpip_pipe2socket.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("ERROR: dkg-tcpip-common (close)");
+			perror("WARNING: dkg-tcpip-common (close)");
 	}
 	for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket.begin();
 		pi != tcpip_broadcast_pipe2socket.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("ERROR: dkg-tcpip-common (close)");
+			perror("WARNING: dkg-tcpip-common (close)");
 	}
 }
 
@@ -1051,24 +1132,35 @@ void tcpip_done
 		if (opt_verbose)
 			std::cerr << "INFO: kill(" << thispid << ", SIGTERM)" << std::endl;
 		if (kill(thispid, SIGTERM))
-			perror("ERROR: dkg-tcpip-common (kill)");
+			perror("WARNING: dkg-tcpip-common (kill)");
 		if (opt_verbose)
+		{
 			std::cerr << "INFO: waitpid(" << thispid << ", NULL, 0)" <<
 				std::endl;
+		}
 		if (waitpid(thispid, NULL, 0) != thispid)
-			perror("ERROR: dkg-tcpip-common (waitpid)");
+			perror("WARNING: dkg-tcpip-common (waitpid)");
 	}
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-tcpip-common (close)");
+				perror("WARNING: dkg-tcpip-common (close)");
 			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
 				(close(broadcast_pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-tcpip-common (close)");
+			{
+				perror("WARNING: dkg-tcpip-common (close)");
+			}
 		}
 	}
+	if ((close(self_pipefd[0]) < 0) || (close(self_pipefd[1]) < 0))
+		perror("WARNING: dots-tcpip-common (close)");
+	if ((close(broadcast_self_pipefd[0]) < 0) ||
+		(close(broadcast_self_pipefd[1]) < 0))
+	{
+		perror("WARNING: dots-tcpip-common (close)");
+	}
 }
 
 int run_tcpip
